/* 
 *  simple audio Lib .. 
 *  copyrights (c) 1994,1995,1996 by Michael Hipp
 *  Replaced with stubs for the Creative Labs derived code use under DOS
 *  
 *  copy policy: GPL V2.0
 */

#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

#include "mpg123.h"
#include "mlib.h"

static int currentRate=44100, currentBits=16, currentChannels=2;
static int initDone=0;

int nBusy;
FILE *rawFile=NULL;

int audio_open(struct audio_info_struct *ai)
{ int err;
  if (rawFile)
  { initDone=1;
	return 0;
  }
  if (initDone)
  { audio_close(ai);
  }
  
  err=audioInit(currentRate, currentBits, currentChannels);
  initDone=(!err);
  return err;
}

int audio_set_rate(struct audio_info_struct *ai)
{
  if (rawFile)
  { return 0;
  }
  if (ai->rate!=currentRate)
  { audio_close(ai);
    currentRate=ai->rate;
  }
  if (!initDone)
  { return audio_open(ai);
  }
  return 0;
}

int audio_set_channels(struct audio_info_struct *ai)
{
  if (rawFile)
  { return 0;
  }
  if (ai->channels!=currentRate)
  { audio_close(ai);
    currentChannels=ai->channels;
  }
  if (!initDone)
  { return audio_open(ai);
  }
  return 0;
}

int audio_play_samples(struct audio_info_struct *ai,short *buf,int len)
{ int totalBytes,nBytes;
  if (rawFile!=NULL)
  {	return fwrite((char *)buf,len,sizeof(short),rawFile)*sizeof(short);
  }

  len*=2; /* Convert number of 16-bit samples to number of bytes */
  totalBytes=0;
  while (len>0)
  { nBytes=audioPlay(((char *)buf)+totalBytes,len);
/* printf("%5lX %5lX %5lX\n",totalBytes, len, nBytes); */
    totalBytes+=nBytes;
    len-=nBytes;
    if (len==0)
    { nBusy++;
      break;
    } else
    { /* Didn't get them all in ! */
      run_idle_task();
    }
  }
  return totalBytes;
}

void audio_clear(void)
{ 
  if (rawFile)
  { return;
  }
  audioComplete();
}

int audio_close(struct audio_info_struct *ai)
{
  if (rawFile)
  { initDone=0;
	return 0;
  }
  if (initDone)
  { audioExit();
  }
  initDone=0;
  return 0;
}
