/* 
 * Com port or local port display driver
 * -------------------------------------
 * copyright (c) 1998-1999 by Roger Walker, All rights reserved.
 * See the file RNW_COPY
 *
 * Uses the routines in serial.c for the com port support
 *
 *  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
 *  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
 *  PURPOSE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <dos.h>
#include <ctype.h>
#include <memory.h>
#include "jukebox.h"

// Single character codes
#define VT_CH_BELL	'\007'		// (7)	Beeps the terminal.
#define VT_CH_BS	'\010'		// (8)	Moves the cursor back one column.
#define VT_CH_HT	'\011'		// (9)	Moves the cursor to the next tab stop (terminal dependant).
#define VT_CH_LF	'\012'		// (10)	Moves the cursor down one row.
#define VT_CH_CR	'\015'		// (13)	Moves the cursor to column one.
#define VT_CH_CAN	'\030'		// (24) Cancels an Escape sequence.
#define	VT_CH_ESC	'\033'		// (27) Starts an Escape sequence.

// NULL terminated version
#define	VT_ESC		"\033"		// (27) Starts an Escape sequence.

// Multi byte escape sequences
#define VT_52		VT_ESC"[?2l"	// Select VT52 mode from VT100 mode
#define VT_100		VT_ESC"<"		// Select VT100 mode from VT52 mode
#define VT_UP		VT_ESC"A"		// Cursor up
#define VT_DOWN		VT_ESC"B"		// Cursor down
#define VT_RIGHT	VT_ESC"C"		// Cursor right
#define VT_LEFT		VT_ESC"D"		// Cursor left
#define VT_GRAPH	VT_ESC"F"		// Special graphics character set
#define VT_TEXT		VT_ESC"G"		// Select ASCII character set
#define VT_HOME		VT_ESC"H"		// Cursor to home
#define VT_REVLF	VT_ESC"I"		// Reverse line feed
#define VT_CLES		VT_ESC"J"		// Erase to end of screen
#define VT_CLEL		VT_ESC"K"		// Erase to end of line
#define VT_XY		VT_ESC"Y"		// Direct cursor address + line+32d + column+32d, 0 based
#define VT_ID		VT_ESC"Z"		// Identify (returns ESC/Z)
#define VT_NUM_ON	VT_ESC"="		// Enter alternate keypad mode
#define VT_NUM_OFF	VT_ESC">"		// Exit alternate keypad mode
#define VT_GR_ON	VT_ESC"1"		// Graphics processor on (only if graphics proc installed)
#define VT_GR_OFF	VT_ESC"2"		// Graphics processor off (only if graphics proc installed)


// These codes are used for my custom terminal software
#define Z_LEFT		('H'-'@')		// Cursor left
#define Z_RIGHT		('I'-'@')		// Cursor right
#define Z_DOWN		('J'-'@')		// Cursor down
#define Z_UP		('K'-'@')		// Cursor up
#define Z_CLS		('L'-'@')		// Erase to end of screen
#define Z_CR		('M'-'@')		// Cursor to first column of next line
#define Z_HOME		('N'-'@')		// Cursor to home
#define Z_CLEL		('O'-'@')		// Erase to end of line
#define Z_ID		('X'-'@')		// Same as VT_CH_CAN so that ID code is easy, this is echo'd by receiver
#define Z_ROW		0xC0			// These two bits set for a row set
#define Z_COL		0x80			// This bit set for a column set

static enum { io_vt52, io_custom, io_crt } io_mode=io_crt;
static int last_x,last_y;
static WORD last_pos;

int display_height=10;

// Character attribute for white text on black background
#define WHITE_ON_BLACK 7

// Max size of screen
#define SCREEN_WIDTH  80
#define SCREEN_HEIGHT 25


// Do a DOS int 10h (video software interrupt) with selected registers
static WORD doint10(int ax,int bx,int cx,int dx)
{ union REGS rio;
  rio.x.ax=ax; rio.x.bx=bx; rio.x.cx=cx; rio.x.dx=dx;
  int86(0x10,&rio,&rio);
  return rio.x.dx;
}


// Send a string to the currently selected output device
void display_str(char *str)
{ char ch;
  while ((ch= *(str++))!='\0') 
   switch (io_mode)
	{ case io_crt:
	    if ((last_pos&0xFF00)==0xFF00) break;
		if (ch=='\b')
		{ if ((last_pos&0xFF)!=0) last_pos--;
		} else
		if (ch=='\n')
		{ last_pos&=0xFF00; 
		  last_pos+=0x0100;
		  if (last_pos>=(SCREEN_HEIGHT<<8)) 
		  { doint10(0x0601,WHITE_ON_BLACK<<8,0,((SCREEN_HEIGHT-1)<<8)|(SCREEN_WIDTH-1));
		    last_pos-=0x0100;
		  }
		} else
		{ doint10(0x0900|ch,WHITE_ON_BLACK,1,0);
		  last_pos++;
		}
		doint10(0x0200,0,0,last_pos);
		break;
	
	  case io_vt52:
		if (ch=='\b')
		{ serial_tx(VT_CH_BS);
		  last_x--;
		} else
		if (ch=='\n')
		{ serial_tx(VT_CH_LF);
	      serial_tx(VT_CH_CR);
		  last_y= -1;
		} else
		{ serial_tx(ch);
		  last_x++;
		}
		break;

	  case io_custom:
		if (ch=='\b')
		{ serial_tx(Z_LEFT);
		  last_x--;
		} else
		if (ch=='\n')
		{ serial_tx(Z_CR);
		  last_y= -1;
		} else
		{ serial_tx(ch);
		  last_x++;
		}
		break;
	}
}


// Wait for the currently sent characters to actually be sent (mainly for RS232)
void display_flush(void)
{ if (io_mode==io_crt) return;
  while (!serial_tx_empty())
  { // Wait for TX to finish sending or for a character to be received
  }
}


// Initialise the output device (and try to detect an RS232 terminal)
int display_init(void)
{ int retry,timeout,chgot;
  
  retry=serial_init();
  if (retry) return retry;

  for (retry=0 ; retry<3 ; retry++)
  { // In case we stopped in the middle of an escape sequence
	// It's also echo'd by the stateless terminal as Z_ID
    serial_tx(VT_CH_CAN);
	// Clear input buffer
	timeout=0;
    while (serial_rx_ready()) 
	{ timeout=serial_rx();
	  if (timeout==Z_ID)
	  { io_mode=io_custom;
	    display_height=8;
        fprintf(stderr,"Connected to custom terminal\n");
	    return 0;
	  }
	  fprintf(stderr,"0x%02X ",timeout);
	}

    // Force VT52 mode, send ID command
    io_mode=io_vt52;
    display_str("   " VT_52 "   " VT_ID);

	for (timeout=0 ; timeout<20 ; timeout++)
	{ serial_tx(' ');

      while (!serial_tx_empty())
	  { // Wait for TX to finish sending or for a character to be received
	    if (serial_rx_ready()) break;
	  }
	  
	  // Check if it's the returned esc sequence
	  if (serial_rx_ready())
	  { chgot=serial_rx();
	    printf("a%02X ",chgot);
		if (chgot!=VT_CH_ESC) continue;
		
		chgot=serial_rx();
		printf("b%02X ",chgot);
        if (chgot!='/') continue;
		
		chgot=serial_rx();
		printf("c%02X ",chgot);
		if (chgot!='Z') continue;

		io_mode=io_vt52;
	    display_height=24;
		fprintf(stderr,"Connected to VT52 terminal\n");
		return 0;
	  }
	}
  }
  
  // Didn't find an RS232 terminal, so default to local keyboard/VDU
  io_mode=io_crt;
  display_clear();
  display_height=SCREEN_HEIGHT-2;
  display_str("Using VGA display and keyboard");
  return 0;
}


// Cancel the current output queue, (only really for RS232 devices)
// Returns true if it actually truncated the output
bool display_cancel(void)
{ // Only needed by serial terminals
  if (io_mode==io_crt) return false;

  // Check if TX queue is already empty
  if (serial_tx_empty()) return false;
  
  // Clear it
  serial_tx_clear();	
  // Then send cancel code in case we were in the middle of an escape sequence
  if (io_mode==io_vt52) 
  { display_str(VT_HOME VT_HOME);
    display_flush();
  } else
  { serial_tx(Z_HOME);
  }
  last_x=last_y=0;
  return true;
}


// Clear the screen, and put the cursor in the top left
void display_clear(void)
{ if (io_mode==io_crt)
  { doint10(0x0002,0,0,0);
    last_pos=0;
  } else
  { display_cancel();
    if (io_mode==io_vt52)
	{ display_str("   " VT_52 VT_HOME VT_CLES VT_HOME);
	  display_flush();
	} else
	{ serial_tx(Z_CLS);
	}
    last_x=last_y=0;
  }
}


// Close down the output device
void display_exit(void)
{ display_clear();
  display_str("Goodbye");

  // Serial modes wait for the string to be sent before exiting
  if (io_mode!=io_crt)
  { while (!serial_tx_empty())
	{ // Wait for characters to be sent
	}
  }

  // Even the CRT inits the serial port
  serial_exit();
}


// Move the cursor position to the selected position, 0,0 = Top,left
void display_xy(int col,int row)
{ 
  if (io_mode==io_crt)
  { last_pos=(row<<8)|col;
    doint10(0x0200,0,0,last_pos);
	return;
  }

  if (last_x==col && last_y==row) return;
  if (io_mode==io_vt52)
  { display_str(VT_XY);
    serial_tx(' '+row);
    serial_tx(' '+col);
  } else
  { serial_tx(Z_ROW|row);
    serial_tx(Z_COL|col);
  }
  last_x=col; last_y=row;
}


// Move the cursor to the start of the selected row, 0=top
void display_row(int row)
{ 
  if (io_mode==io_crt)
  { display_xy(0,row);
    return;
  }
 
  if (last_x==0 && last_y==row) return;
  if (io_mode==io_vt52)
  { display_xy(0,row);
  } else
  { if (last_x!=0 || last_y!=row)
      serial_tx(Z_ROW|row);
  }
  last_x=0; last_y=row;
}

// Clear from the current cursor position to the end of the line
void display_clr_eol(void)
{ if (io_mode==io_crt)
  { if ( (last_pos&0xFF00)!=0xFF00 && (last_pos&0xFF)<(SCREEN_WIDTH-1))
	{ doint10(0x0920,WHITE_ON_BLACK,(SCREEN_WIDTH-1)-(last_pos&0xFF),0);
      last_pos=0xFF00|SCREEN_WIDTH;
	}
  } else
  if (io_mode==io_vt52)
  { display_str(VT_CLEL);
  } else
  { serial_tx(Z_CLEL);
  }
}

/* VT52 key		TRANSMITS
 *
 * up-arrow		ESC A
 * down-arrow	ESC B
 * right-arrow	ESC C
 * left-arrow	ESC D
 *
 * PF1 key		ESC P
 * PF2 key		ESC Q
 * F3			ESC 0x3F 0x77
 * F4			ESC 0x3F 0x78
 * F5			ESC 0x3F 0x74
 * F6			ESC 0x3F 0x75
 * F7			ESC 0x3F 0x71
 * F8			ESC 0x3F 0x72
 * F9			ESC 0x3F 0x70
 */

// If no key ready return 0, otherwise return PC keycode
int display_getkey(void)
{ static int got_esc=0;
  int chgot;

  while (1)
   { // Anything in the serial input buffer ?
     if (!serial_rx_ready()) 
	 { // Nothing in RX buffer, check PC keyboard buffer
	   if (!kbhit()) 
	   { // Nothing there either
		 return 0;
	   }

	   // Get key from PC buffer
	   chgot=getch();
	   if (chgot==0)
	   { // Return extended keys as 256+key value
	     chgot=256+getch();
	   }
	   return chgot;
	 }

	 // Get the waiting character
	 chgot=serial_rx();

	 // If we are in serial mode, then echo the serial input sream for debugging purposes
	 printf("%d%d%02X ",io_mode,got_esc,chgot);

	 // CAN aborts an ESC sequence, otherwise is ignored
	 if (chgot==24)
	 { got_esc=0;
	   continue;
	 }

	 // Either ESC, or ESC 0x3F
	 if (chgot==27 || (got_esc==1 && chgot==0x3F))
	 { got_esc++;
	   continue;
	 }

	 if (got_esc==1 && (chgot==0x4F || chgot==0x5B))
	 { // VT100 mode !
	   got_esc=3;
	   continue;
	 }

	 // Third character in VT100 sequence
	 if (got_esc==3)
	 { // Just lose character, then put terminal into VT52 mode
	   got_esc=0;
	   io_mode=io_vt52;
	   display_str("   " VT_52);
	   continue;
	 }

	 // Not yet in the middle of an escape sequence
	 if (got_esc==0)
	 { 
	   switch(chgot)
	   {	 case 3:   return KEY_CF12; // Return control+C as CTRL_F12

		 case 8:
		 case 127: return KEY_DEL;  // either backspace or del key as delete

		 case 10:
		 case 13:  return KEY_RET;  // either linefeed or CR as return key

	   
		 case 0x80|'1': io_mode=io_custom; return KEY_F1;
		 case 0x80|'2': io_mode=io_custom; return KEY_F2;
		 case 0x80|'3': io_mode=io_custom; return KEY_F3;
		 case 0x80|'4': io_mode=io_custom; return KEY_F4;
		 case 0x80|'5': io_mode=io_custom; return KEY_F5;
		 case 0x80|'6': io_mode=io_custom; return KEY_F6;
		 case 0x80|'7': io_mode=io_custom; return KEY_F7;
		 case 0x80|'8': io_mode=io_custom; return KEY_F8;
		 case 0x80|'u': io_mode=io_custom; return KEY_UP;
		 case 0x80|'d': io_mode=io_custom; return KEY_DOWN;
		 case 0x80|'U': io_mode=io_custom; return KEY_PGUP;
		 case 0x80|'D': io_mode=io_custom; return KEY_PGDN;
	   }
	  return chgot;
	 }

	 // Received 1 character from the escape sequence
	 if (got_esc==1)
	 { // We previously received an ESC, check if this is a sequence we recognise
	   got_esc=0;
	   switch (chgot)
	   { case 'A': return KEY_UP;
	     case 'B': return KEY_DOWN;
	     case 'C': return KEY_PGDN;
	     case 'D': return KEY_PGUP;

	     case 'P': return KEY_F1;
	     case 'Q': return KEY_F2;

	     case 'R': return KEY_F3;
	     case 'S': return KEY_F4;
	   }
	   continue;
	 }

	 // Received 2 characters from the escape sequence
	 if (got_esc==2)
	 { // ESC 0x3F <n>
	   got_esc=0;
	   switch (chgot)
	   {	case 0x77: return KEY_F3; break;
		case 0x78: return KEY_F4; break;
		case 0x74: return KEY_F5; break;
		case 0x75: return KEY_F6; break;
		case 0x71: return KEY_F7; break;
		case 0x72: return KEY_F8; break;
		case 0x70: return KEY_F9; break;
	   }
	   continue;
	 }
   }
}
