/* 
 * local display driver
 * -------------------------------------
 * copyright (c) 1998-1999 by Roger Walker, All rights reserved.
 * See the file RNW_COPY
 *
 * Cut down version of disp_com.c
 *
 *  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
 *  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
 *  PURPOSE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <dos.h>
#include <ctype.h>
#include <memory.h>
#include "jukebox.h"

// Character attribute for white text on black background
#define WHITE_ON_BLACK 7

// Max size of screen
#define SCREEN_WIDTH  80
#define SCREEN_HEIGHT 25

int display_height=(SCREEN_HEIGHT-1);

// Do a DOS int 10h (video software interrupt) with selected registers
static WORD doint10(int ax,int bx,int cx,int dx)
{ union REGS rio;
  rio.x.ax=ax; rio.x.bx=bx; rio.x.cx=cx; rio.x.dx=dx;
  int86(0x10,&rio,&rio);
  return rio.x.dx;
}

// The current cursor position
static WORD pos;


// Initialise the output device
int display_init(void)
{ doint10(0x0002,0,0,0);
  pos=0;
  return 0;
}


// Move the cursor position to the selected position, 0,0 = Top,left
void display_xy(int x,int y)
{ pos=(y<<8)|x;
  doint10(0x0200,0,0,pos);
}


// Move the cursor to the start of the selected row, 0=top
void display_row(int row)
{ display_xy(0,row);
}


// Clear the screen, and put the cursor in the top left
void display_clear(void)
{ display_init();
}


// Cancel the current output queue, (only really for RS232 devices)
// Returns true if it actually truncated the output
bool display_cancel(void)
{ return false;
}


// Wait for the currently sent characters to actually be sent (only really for RS232 devices)
void display_flush(void)
{
}


// Display a single character, updating the cursor position
static void display_ch(char ch)
{ if (ch=='\b')
  { if ((pos&0xFF)!=0) pos--;
  } else
  if (ch=='\n')
  { pos&=0xFF00; 
    pos+=0x0100;
    if (pos>=(SCREEN_HEIGHT<<8)) 
	{ doint10(0x0601,WHITE_ON_BLACK<<8,0,((SCREEN_HEIGHT-1)<<8)|(SCREEN_WIDTH-1));
	  pos-=0x0100;
	}
  } else
  { doint10(0x0900|ch,WHITE_ON_BLACK,1,0);
    pos++;
  }
  doint10(0x0200,0,0,pos);
}

// Send a string to the currently selected output device
void display_str(char *str)
{ while (*str!='\0') display_ch( *(str++) );
}

// Close down the output device
void display_exit(void)
{ display_clear();
  display_str("Goodbye");
}

// Clear from the current cursor position to the end of the line
void display_clr_eol(void)
{ if ((pos&0xFF)<(SCREEN_WIDTH-1))
  { doint10(0x0920,WHITE_ON_BLACK,SCREEN_WIDTH-(pos&0xFF),0);
    pos=(pos&0xFF00)|(SCREEN_WIDTH-1);
	doint10(0x0200,0,0,pos);
  }
}

// If no key ready return 0, otherwise return PC keycode
int display_getkey(void)
{ int chgot;

  if (!kbhit()) 
  { 
	  return 0;
  }

  
  chgot=getch();
  if (chgot==0)
  { chgot=256+getch();
  }
  return chgot;
}


#ifdef BUILD_TEST
// Simple test program
main()
{ int i,j; char buf[64];
  i=1024;
  for (j=0 ; j<24 ; j++)
  { display_xy(j,j);
    sprintf(buf,"%d",i++); display_str(buf);
  }
  for (j=0 ; j<24 ; j++)
  { display_xy(70-j,j);
    sprintf(buf,"%d abcdefghijk",i++); display_str(buf);
  }
  getch();
  for (j=0 ; j<24 ; j++)
  { display_xy(2+j,j);
    display_clr_eol();
  }
}
#endif
