/* 
 * Com port or local port display driver
 * -------------------------------------
 * copyright (c) 1998-1999 by Roger Walker, All rights reserved.
 * See the file RNW_COPY
 *
 * Uses the routines in serial.c for the com port support
 *
 *  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
 *  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
 *  PURPOSE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <dos.h>
#include <ctype.h>
#include <memory.h>
#include "jukebox.h"
#include "disp_win.h"
//#include "mlib.h"

// Max size of screen
#define SCREEN_WIDTH  80
#define SCREEN_HEIGHT 25

window_t windows[4]=
{  0, 0,80, 4,2,2,	// Rows  0-3  Current playing status + What's lined up   
   0, 5,80,17,0,3,	// Rows  6-21 Main window
   5, 7,70,13,0,2,	// Rows  7-20 Pop up window (Completely enclosed in the main window)
   0,23,80, 2,0,1	// Rows 23-24 Error window
};

// Row 5 = Line, 21 = Line

window_t *pCurrentWin=NULL;
int currentWinNum;

// Character attributes
#define WHITE_ON_BLACK 15
#define BLACK_ON_GREY  (7<<4)
#define GREY_ON_BLACK  7

static WORD screen_pos;
static BYTE screen_attr=WHITE_ON_BLACK;

static void doint10(int ax,int bx,int dx)
{ union REGS rio;
  rio.x.ax=ax; rio.x.bx=bx; rio.x.cx=0; rio.x.dx=dx;
  int86(0x10,&rio,&rio);
}

static void screen_move_newpos(unsigned short pos)
{ screen_pos=pos;
  doint10(0x0200,0,screen_pos);
}

static void screen_move_xy(int x,int y)
{ screen_move_newpos(y*256+x);
}

static void screen_clear(void)
{ doint10(0x600,WHITE_ON_BLACK<<8,24*256+79);
  doint10(0x0500,0,0);
  screen_move_xy(0,0);
}

static void screen_clear_n(int nch)
{ union REGS rio;
  if (nch>0)
  { rio.x.ax=(9<<8)|' '; rio.x.bx=screen_attr; rio.x.cx=nch;
    int86(0x10,&rio,&rio);
	screen_pos+=nch;
	screen_move_newpos(screen_pos);
  }
}

static void screen_print_n(char *line,int maxwidth)
{ int chlp; union REGS rin,rout;
  rin.h.ah=9; rin.x.bx=screen_attr; rin.x.cx=1;
  chlp=0;
  while (line[chlp] && chlp<maxwidth)
   { rin.h.al=line[chlp++]; int86(0x10,&rin,&rout);
     screen_move_newpos(++screen_pos);
   }
  if (chlp<maxwidth) screen_clear_n(maxwidth-chlp);
}

void window_clear(void)
{ int row;
  for (row=0 ; row<pCurrentWin->height ; row++)
  { screen_move_xy(pCurrentWin->left,pCurrentWin->top+row);
    screen_clear_n(pCurrentWin->width);
  }
  screen_move_xy(pCurrentWin->left,pCurrentWin->top);
  pCurrentWin->cur_pos=screen_pos;
}

void window_move_xy(int x,int y)
{ screen_move_xy(pCurrentWin->left+x,pCurrentWin->top+y);
  pCurrentWin->cur_pos=screen_pos;
}

void window_clear_n(int maxWidth)
{ if (pCurrentWin->cur_pos!=screen_pos) screen_move_newpos(pCurrentWin->cur_pos);
  if (((pCurrentWin->cur_pos&0xFF)+maxWidth)>(pCurrentWin->left+pCurrentWin->width))
  { maxWidth=pCurrentWin->left+pCurrentWin->width-(pCurrentWin->cur_pos&0xFF);
  }
  if (maxWidth>0)
  {	screen_clear_n(maxWidth);
	pCurrentWin->cur_pos=screen_pos;
  }
}

void window_print_n(char *line,int maxWidth)
{ if (pCurrentWin->cur_pos!=screen_pos) screen_move_newpos(pCurrentWin->cur_pos);
  if (((pCurrentWin->cur_pos&0xFF)+maxWidth)>(pCurrentWin->left+pCurrentWin->width))
  { maxWidth=pCurrentWin->left+pCurrentWin->width-(pCurrentWin->cur_pos&0xFF);
  }
  if (maxWidth>0)
  {	screen_print_n(line,maxWidth);
	pCurrentWin->cur_pos=screen_pos;
  }
}

void window_print(char *line)
{ window_print_n(line,min(strlen(line),pCurrentWin->width-(pCurrentWin->cur_pos)&0xFF));
}

void window_print_clr(char *line)
{ window_print_n(line,pCurrentWin->width);
}

void window_hilite(int on)
{ screen_attr=(on==1) ? BLACK_ON_GREY : (on==2) ? GREY_ON_BLACK : WHITE_ON_BLACK;
}

void window_border(void)
{ int i;

  window_hilite(2);

  // If this window is full width then only draw the top and bottom lines
  if (pCurrentWin->left==0 && pCurrentWin->width==SCREEN_WIDTH)
  { if (pCurrentWin->top>0)
	{ screen_move_xy(pCurrentWin->left,pCurrentWin->top-1);
	  for (i=0 ; i<pCurrentWin->width ; i++)
		screen_print_n("\xCD",1);
	}
	if ((pCurrentWin->top+pCurrentWin->height)>0)
	{ screen_move_xy(pCurrentWin->left,pCurrentWin->top+pCurrentWin->height);
	  for (i=0 ; i<pCurrentWin->width ; i++)
		screen_print_n("\xCD",1);
	}
	return;
  }

  screen_move_xy(pCurrentWin->left-1,pCurrentWin->top-1);
  screen_print_n("\xC9",1);
  for (i=0 ; i<pCurrentWin->width ; i++)
	  screen_print_n("\xCD",1);
  screen_print_n("\xBB",1);

  screen_move_xy(pCurrentWin->left-1,pCurrentWin->top+pCurrentWin->height);
  screen_print_n("\xC8",1);
  for (i=0 ; i<pCurrentWin->width ; i++)
	  screen_print_n("\xCD",1);
  screen_print_n("\xBC",1);

  for (i=0 ; i<pCurrentWin->height ; i++)
  {	screen_move_xy(pCurrentWin->left-1,pCurrentWin->top+i);
	screen_print_n("\xBA",1);
	screen_move_xy(pCurrentWin->left+pCurrentWin->width,pCurrentWin->top+i);
	screen_print_n("\xBA",1);
  }
  window_hilite(0);
}

int window_select(int winNum)
{ int lastWin;
  lastWin=currentWinNum;
  currentWinNum=winNum;
  pCurrentWin=windows+winNum;
  if (pCurrentWin->cur_pos!=screen_pos) screen_move_newpos(pCurrentWin->cur_pos);
  return lastWin;
}

// Send a string to the currently selected output device
void display_str(char *str)
{ int lastWinNum;

  // Save last window
  lastWinNum=currentWinNum;

  window_select(WIN_ERR);
  window_print(str);
  window_select(lastWinNum);
}


// Wait for the currently sent characters to actually be sent (mainly for RS232)
void display_flush(void)
{
}


// Initialise the output device
int display_init(void)
{ int i;
  display_clear();
  for (i=0 ; i<4 ; i++)
  { window_select(i);
    window_move_xy(0,0);    
  }  
  return 0;
}


// Cancel the current output queue, (only really for RS232 devices)
// Returns true if it actually truncated the output
bool display_cancel(void)
{
  return true;
}


// Clear the screen, and put the cursor in the top left
void display_clear(void)
{ union REGS rio;
  doint10(0x0002,0,0);

  // AH = 1
  // CH bits 5-6 = blink attribute (00=normal, 01=invisible, 10 = slow, 11=fast)
  // CH bits 0-4 = start line for cursor in char. cell
  // CL bits 0-4 = end line for cursor in char. cell
  rio.h.ah=1;
  rio.x.bx=rio.x.dx=0;
  rio.x.cx=0x4A09;
  int86(0x10,&rio,&rio);
}

// Output a character at a specific x,y position
void display_char(int x,int y,char ch)
{ screen_move_xy(x,y);
  screen_print_n(&ch,1);
}


// Close down the output device
void display_exit(void)
{ display_clear();
  display_str("Goodbye");
}


// If no key ready return 0, otherwise return PC keycode
int display_getkey(void)
{ int chgot;

  // Nothing in the keyboard buffer so return 0  
  if (!kbhit()) return 0;

  // Get key from PC buffer
  chgot=getch(); 
  // Return extended keys as 256+key value
  if (chgot==0) chgot=256+getch();

  return chgot;
}
