/* 
 * Find the mp3 files to include in the play list
 * ----------------------------------------------
 * copyright (c) 1998-1999 by Roger Walker, All rights reserved.
 * See the file RNW_COPY
 *
 * Looks for a "wavindex.txt" in the root directories of all hard drives
 *
 *  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
 *  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
 *  PURPOSE.
 */

#include <stdio.h>
#include <dos.h>
#include <conio.h>
#include <ctype.h>
#include <memory.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <search.h>
#include "jukebox.h"

play_list_t toplay[MAX_PLAY];
int ntoplay,max_toplay,total_played;
bool really_exit,background_mode;

/* These are for access by external routines */
trk_info_t *trks;
int ntrks,nsorted;
unsigned tracks_bya[MAX_LOAD], tracks_byt[MAX_LOAD];
trk_list_t trk_lists[MAX_STYLE];

// String buffers
static char **strbuf;
static unsigned nstr;
static char *bad_file=NULL;

#define BUF_BLK_SIZ 1023
// Grow a string buffer to allow for an extra entry
static void grow_buffer(void **buf,unsigned int newsize)
{ if (*buf==NULL) *buf=malloc(newsize|BUF_BLK_SIZ);
  else *buf=realloc(*buf,newsize|BUF_BLK_SIZ);
  if (*buf==NULL)
   { printf("Failed to grow buffer to %u bytes\n",newsize);
     exit(0);
   }
}


#define MAX_SBUF 8
#define SBUF_SIZE 65535U
unsigned str_buf_lens[MAX_SBUF];
char *str_bufs[MAX_SBUF];

static char *str_malloc(unsigned len)
{ unsigned i;
  char *cp;
  cp=NULL;
  for (i=0 ; i<MAX_SBUF ; i++)
  if ((SBUF_SIZE-str_buf_lens[i])>=len)
  { if (str_buf_lens[i]==0)
    { str_bufs[i]=malloc(SBUF_SIZE);
      if (str_bufs[i]==NULL) return NULL;
    }
    cp=str_bufs[i]+str_buf_lens[i];
    str_buf_lens[i]+=len;
    break;
  }
  return cp;
}

// Return the stored string pointer for a string
// (If the string doesn't exist in list then add it first)
static char *make_string(char *prefix,char *buf,unsigned len)
{ char *ret;
  unsigned i,pLen;
  pLen=strlen(prefix);
  for (i=0 ; i<nstr ; i++)
   if ((pLen==0 || strnicmp(strbuf[i],prefix,pLen)==0) &&
       (len==0 || strnicmp(strbuf[i]+pLen,buf,len)==0) &&
       strbuf[i][pLen+len]=='\0')
     return strbuf[i];
  grow_buffer((void **)&strbuf,(nstr+1)*sizeof(char *));
  ret=str_malloc(pLen+len+1);
  if (ret!=NULL)
   { strcpy(ret,prefix);
     strncpy(ret+strlen(prefix),buf,len);
     ret[strlen(prefix)+len]='\0';
     strbuf[nstr++]=ret;
   }
  return ret;
}


// Sort the track list
static int _cdecl sort_trks(void *p1, void *p2)
{ int val;
  val=stricmp( ((trk_info_t *)p1)->artist, ((trk_info_t *)p2)->artist);
  if (val) return val;
  return stricmp( ((trk_info_t *)p1)->title, ((trk_info_t *)p2)->title);
}


// Sort the artist list
static int _cdecl sort_by_artist(const void *vp1,const void *vp2)
{ trk_info_t *p1,*p2; int val; char *cp1,*cp2;
  p1=trks+(*(int *)vp1); p2=trks+(*(int *)vp2);
  cp1=p1->artist; if (strnicmp(cp1,"the ",4)==0) cp1+=4;
  cp2=p2->artist; if (strnicmp(cp2,"the ",4)==0) cp2+=4;
  val=stricmp(cp1, cp2);
  if (val) return val;
  return stricmp(p1->title,p2->title);
}


// Sort the play list by title
static int _cdecl sort_by_title(const void *vp1, const void *vp2)
{ trk_info_t *p1,*p2; int val;
  p1=trks+(*(int *)vp1); p2=trks+(*(int *)vp2);
  val=stricmp(p1->title,p2->title);
  if (val) return val;
  return stricmp(p1->artist, p2->artist);
}


// Sort the play list by score
static int _cdecl sort_tracks(const void *v1,const void *v2)
{ play_list_t *p1,*p2;
  p1=(play_list_t *)v1;
  p2=(play_list_t *)v2;

  // Lowest sequence number is best
  if (p1->sequence!=p2->sequence)
    return (p1->sequence<p2->sequence) ? -1 : 1;

  // Highest score is best
  if (p1->score!=p2->score)
    return (p1->score>p2->score) ? -1 : 1;

  // Lowest time is best
  return (p1->time<p2->time) ? -1 : 
         (p1->time>p2->time) ? 1 : 0;
}


// Compress multiple whitespace to a single space and remove trailing control codes and spaces
static void tidy_line(char *buf)
{ char *cp;
  /* Replace multiple whitespace with a single space */
  cp=buf;
  while (cp[0]!='\0' && cp[1]!='\0')
  { if (isspace(cp[0]))
    { cp[0]=' ';
      while (isspace(cp[1]))
      { strcpy(cp+1,cp+2);
      }
    }
    cp++;
  }

  /* Remove control codes and spaces from end of line */
  cp=buf+strlen(buf);
  while (cp>buf && *cp=='\0')
  { cp--;
    if (iscntrl(*cp) || isspace(*cp)) *cp='\0';
  }
}


// Skip backwards over blank tracks 
// (tracks prefixed with @@ are used as place holders in wavindex.txt files)
unsigned track_valid(unsigned trk)
{ return (trks[trk].fname!=bad_file) ? trk : 
         (trk>0) ? track_valid(trk-1) : 0;
}


// Returns true if the file exists, caches directories
static char *cached_dir=NULL;
static int ncached;

static bool file_exists(char *file_name)
{	char *pDir,*pFname,tofind[512];
	int i;

	if ((pDir=strrchr(file_name,'\\'))!=NULL) 
	{ pDir++;
	} else
	if ((pDir=strrchr(file_name,'/'))!=NULL) 
	{ pDir++;
	} else
	if ((pDir=strrchr(file_name,':'))!=NULL) 
	{ pDir++;
	} else
	{ pDir=file_name;
	}

	if (cached_dir==NULL || strnicmp(file_name,cached_dir,pDir-file_name)!=0)
        {       fprintf(stderr,"Reading %s,",file_name);

                // We don't have a cached dir, or the dir is the wrong one
		if (cached_dir==NULL)
                {       cached_dir=malloc(strlen(file_name)+1024);
		} else
                {       cached_dir=realloc(cached_dir,strlen(file_name)+1024);
		}
		// The malloc or realloc worked and we have enough room to start with
		if (cached_dir!=NULL)
		{	struct find_t found;
			strncpy(cached_dir,file_name,pDir-file_name);
			cached_dir[pDir-file_name]='\0';
			ncached=0;
			strcpy(tofind,cached_dir);
			strcat(tofind,"*.*");
			pFname=cached_dir+strlen(cached_dir)+1;
			if (_dos_findfirst(tofind,_A_NORMAL,&found)==0)
			do	{	char *pNew;
                                        pNew=realloc(cached_dir,pFname-cached_dir+strlen(found.name)+2);
					if (pNew!=cached_dir)
					{	pFname=pNew+(pFname-cached_dir);
						cached_dir=pNew;
					}
					strcpy(pFname,found.name);
					pFname+=strlen(found.name)+1;
					ncached++;
				}
			while (_dos_findnext(&found)==0);
                        fprintf(stderr," %u found\n",ncached);
		}
        }

	if (cached_dir!=NULL)
	{	// We've a cached version of this dir
		pFname=cached_dir;
		for (i=0 ; i<ncached ; i++)
		{	// Entry 0 held the dir name
			pFname+=strlen(pFname)+1;
			if (stricmp(pFname,pDir)==0)
			{	return true;
			}
		}
		// Didn't find it
		return false;
	} else
	return (access(file_name,04)==0);
}


// Scan drives for \wavindex.txt files and read the contents
int track_init(void)
{ unsigned i,ndrives,line,nwarns;
  char drive,*filen="?:\\wavindex.txt",fpath[4],linebuf[512];
  union REGS rio;
  FILE *in;
  int nstyles[MAX_STYLE];
  trk_info_t *tp;

  ntrks=nsorted=0; trks=NULL;

  memset(nstyles,0,sizeof(nstyles));
  nstr=0; strbuf=NULL;

  /* Find the maximum drive letter in the system */
  _dos_getdrive(&i);
  _dos_setdrive(i,&ndrives);

  /* No warnings (yet) */
  nwarns=0;

  /* Clear string buffers */
  memset(str_buf_lens,0,sizeof(str_buf_lens));
  memset(str_bufs,0,sizeof(str_bufs));

  bad_file=make_string("Illegal file","",0);

  for (drive='C' ; drive<((char)('A'+ndrives)) ; drive++)
   { rio.x.ax=0x4409 ; rio.h.bl=(drive-'A'+1) ;
     if (intdos(&rio,&rio)==0x000F) continue ; /* Invalid drive */

     filen[0]=drive;
     /* If index file didn't open then skip drive */
     if ((in=fopen(filen,"rt"))==NULL) continue;

     line=0;
     while (fgets(linebuf,sizeof(linebuf),in))
      { char *cp,*end;
        FILE *test;

        line++;
        tidy_line(linebuf);

        /* If nothing left or comment then skip */
        if (strlen(linebuf)==0 ||
            linebuf[0]=='*' || linebuf[0]=='#' || linebuf[0]==';')
                continue;

        // Found special entry
        if (linebuf[0]=='@' && linebuf[1]!='@')
        { // Found first of the playlist entries so break
          break;
        }

        /* Get space for entry */
        grow_buffer((void **)&trks,(ntrks+1)*sizeof(trk_info_t));

        /* Point to it */
        tp=trks+ntrks;
        memset(tp,0,sizeof(trk_info_t));
        tp->fname=tp->artist=tp->title=NULL;

        // Create blank entry
        if (linebuf[0]=='@' && linebuf[1]=='@')
         { tp->fname=tp->artist=tp->title=bad_file;
           ntrks++;
           continue;
         }

        /* Find end of filename */
        end=strchr(linebuf,' ');
        if (end==NULL)
         { printf("File %s line %d has a bad filename\n'%s'\n",
                filen,line,linebuf);
           nwarns++;
           continue;
         }

        /* Build up full filename with "n:\" at the start */
        /* where n is the drive letter */
        fpath[0]=drive; fpath[1]=':'; 
        cp=fpath+2;
        if (linebuf[0]!='\\') *(cp++)='\\';
        *cp='\0';

        tp->fname=make_string(fpath,linebuf,end-linebuf);

        /* Make sure we can open the file */
		if (!file_exists(tp->fname))
         { printf("File '%s' doesn't exist\n",tp->fname);
           nwarns++;
           continue;
         }

        /* Find first " after the filename */
        cp=strchr(end,'\"');
        if (cp==NULL)
         { printf("File %s line %d has a bad track name (no start \")\n'%s'\n",
                filen,line,linebuf);
           nwarns++;
           continue;
         }
        
        /* Find matching " */
        cp++; end=strchr(cp,'\"');
        if (end==NULL)
         { printf("File %s line %d has a bad track name (no end \")\n'%s'\n",
                filen,line,linebuf);
           nwarns++;
           continue;
         }
        tp->title=make_string("",cp,end-cp);

        /* Find first " after the track */
        cp=strchr(end+1,'\"');
        if (cp==NULL)
         { printf("File %s line %d has a bad artist name (no start \")\n'%s'\n",
                filen,line,linebuf);
           nwarns++;
           continue;
         }
        /* Find matching " */
        cp++; end=strchr(cp,'\"');
        if (end==NULL)
         { printf("File %s line %d has a bad artist name (no end \")\n'%s'\n",
                filen,line,linebuf);
           nwarns++;
           continue;
         }
        tp->artist=make_string("",cp,end-cp);

        /* Check for too many " */
        end=strchr(end+1,'\"');
        if (end!=NULL)
         { printf("File %s line %d has Too many \"s (track included)\n'%s'\n",
                filen,line,linebuf);
           nwarns++;
         }
        ntrks++;
      }

	 // Sort out the play list entries
     if (linebuf[0]=='@')
     { int lastlist= -1;
       char *cp;
       do { tidy_line(linebuf);
            if (strlen(linebuf)==0 ||
               linebuf[0]=='*' || linebuf[0]=='#' || linebuf[0]==';')
                 continue;
            if (linebuf[0]=='@')
            { 
              cp=linebuf+1; 
              while (*cp && !isalpha(*cp)) cp++;
              if (*cp)
              { lastlist=toupper(*cp)-'A';
                if (lastlist>=MAX_STYLE) lastlist= -1;
              }
            } else
            if (lastlist>=0)
            { tidy_line(linebuf);
              cp=strchr(linebuf,'.');
              if (cp!=NULL) *cp='\0';
              for (i=0,tp=trks ; i<(unsigned)ntrks ; i++,tp++)
              { cp=strrchr(tp->fname+strlen(tp->fname)-4,'.');
                if (cp!=NULL &&
                    strnicmp(cp-strlen(linebuf),linebuf,strlen(linebuf))==0) 
                { tp->styles[lastlist]= (++nstyles[lastlist]);
                  break;
                }
              }
              if (i>=(unsigned)ntrks)
              { printf("File %s line %d: File '%s' not found in main list\n",
                    filen,line,linebuf);
                nwarns++;
              }
            }
            line++;
          }
       while (fgets(linebuf,sizeof(linebuf),in));
     }
     fclose(in);
   }


  if (ntrks)
   { /* Generate sorted index lists */
	 if (getenv("SWAPPED")!=NULL)
	 { char *cp;
	   for (i=0 ; i<(unsigned)ntrks ; i++)
		if (trks[i].artist!=bad_file)
		{ cp=trks[i].artist; trks[i].artist=trks[i].title; trks[i].title=cp;
		}
	 }
     for (i=0,nsorted=0 ; i<(unsigned)ntrks ; i++)
	  if (trks[i].artist!=bad_file)
	  { tracks_bya[nsorted]=tracks_byt[nsorted]=i;
	    nsorted++;
	  }

     qsort(tracks_bya,nsorted,sizeof(int),sort_by_artist);
     qsort(tracks_byt,nsorted,sizeof(int),sort_by_title);

	 // Check for identical entries
     for (i=0 ; i<(unsigned)(nsorted-1) ; i++)
      if (trks[tracks_bya[i]].fname!=bad_file &&
          sort_by_artist(tracks_bya+i,tracks_bya+i+1)==0)
       { trk_info_t *p1,*p2;
         p1=trks+tracks_bya[i];
         p2=trks+tracks_bya[i+1];
         printf("Files '%s' and '%s' are both '%s' by '%s'\n",
           p1->fname,p2->fname,p1->title,p1->artist);
         nwarns++;
       }
   }

#if 0
  for (i=0 ; i<(unsigned)ntrks ; i++)
   { printf("%s - %s\n",trks[i].artist,trks[i].title);
   }
  getch();
#endif
  if (nwarns)
   { WORD key;
     printf("\a%d warnings - press a key to continue",nwarns);
     key=display_getkey();
     if (key==KEY_ESC)
      { exit(20);
      }
   }

  return ntrks;
}


// Free memory buffers used by file info
void track_exit(void)
{ 
  if (nstr)
   { int i;
//     for (i=0 ; i<nstr ; i++) 
//       free(strbuf[i]);
     for (i=0 ; i<MAX_SBUF ; i++)
      if (str_buf_lens[i]!=0)
       free(str_bufs[i]);
     free(strbuf);
     nstr=0; strbuf=NULL;
   }

  if (ntrks)
   { free(trks);
     ntrks=0; trks=NULL;
   }
}


// Add a track to somewhere within the play list
int track_add_at(unsigned trk,unsigned pos)
{ char buf[512];
  trk_info_t *tp;

  if (ntoplay>=max_toplay)
   { write_to_log("Attempt to add track when buffer full");
     return -1;
   }

  if (trk>=(unsigned)ntrks || track_valid(trk)!=trk)
  { // Illegal track
    sprintf(buf,"Attempt to add Illegal track %d (of %d) ignored",trk,ntrks);
    write_to_log(buf);
    return -1;
  }

  // Check for double track
  if (background_mode==false && ntoplay>0)
  { int i;
    for (i=0 ; i<ntoplay ; i++)
     if (toplay[i].trk_index==trk)
     { sprintf(buf,"Track %s found in entry %d, track ignored",track_label(trk),i);
       write_to_log(buf);
       return i;
     }
  }

  // If adding before end, move them up by one
  if (((unsigned)ntoplay)>pos) 
  { memmove(toplay+pos+1,toplay+pos,(ntoplay-pos)*sizeof(play_list_t));
  }

  tp=trks+trk;
  sprintf(buf,"Adding %d (%s) as %-3d %-37s%-37s",trk,track_label(trk),ntoplay,tp->artist,tp->title);
  write_to_log(buf);

  time((time_t *)&toplay[pos].time);
  toplay[pos].trk_index=trk;
  toplay[pos].sequence=99999;
  toplay[pos].score=0;
  ntoplay++;
  status_update_tracklist();
  return pos;
}


// Add track to end of play list
int track_add(unsigned trk)
{ return track_add_at(trk,ntoplay);
}


// Calculate the scores of the various tracks then sort them
void track_sort(void)
{ int i;
  unsigned j;
  play_list_t *tp;
  trk_info_t *ti,*fi;
  unsigned long current;
  extern int toplay_status;

  if (ntoplay<=1) return;

  time((time_t *)&current);

  // Find the index for the first track
  fi=trks+toplay[0].trk_index;

  // Work out the scores
  for (i=1,tp=toplay+1 ; i<ntoplay ; i++,tp++)
  { ti=trks+tp->trk_index;
    tp->score=0;
    for (j=0 ; j<MAX_STYLE ; j++)
     if (fi->styles[j]!=0 && ti->styles[j]!=0)
     { tp->score+=20;  // 20 points for each matching style
     }
    // Same artist ?
    if (fi->artist==ti->artist) tp->score+=10;
    // 1 point per minute outstanding
    tp->score+= (current-tp->time)/60;
    // Deduct 25 points if it's been played in the last hour
    if ((current-ti->last_played)<3600) tp->score-=25;
  }

  // Sort the tracks
  qsort(toplay+1,ntoplay-1,sizeof(play_list_t),sort_tracks);
  status_update_tracklist();
}


// Delete the current track
void abort_playing(void)
{ char buf[128];
  if (ntoplay>0)
  { sprintf(buf,"Aborting %s (current track)",track_label(toplay[0].trk_index));
    write_to_log(buf);
  }
  player_abort_playing();
}


// Delete a track from a particular position
void track_delete_from(unsigned pos)
{ if (pos==0)
  { // The tidest way to delete the first track is to abort it
	abort_playing();
  } else
  { char buf[128];
    sprintf(buf,"Deleting %s from position %d of %d",track_label(toplay[pos].trk_index),pos,ntoplay);
    write_to_log(buf);
    ntoplay--;
    if (pos<(unsigned)ntoplay)
        memmove(toplay+pos,toplay+pos+1,(ntoplay-pos)*sizeof(play_list_t));
    status_update_tracklist();
  }
}


// Delete the last entered track
void track_delete_last(void)
{ if (ntoplay>0) track_delete_from(ntoplay-1);
}


// Find a track to play from a play list
int track_find(int block)
{ unsigned *found,nfound,notplayed,i;

  found=malloc(sizeof(int)*ntrks);
  nfound=0;
  for (i=0 ; i<(unsigned)ntrks ; i++)
   if (trks[i].fname!=bad_file)
   { if (block<0)
     { if (trks[i].styles[0]==0) found[nfound++]=i;
     } else
     if (trks[i].styles[block]!=0) found[nfound++]=i;
   }

  // None found in this block so force to 0 or full random
  if (nfound==0) 
  { free(found);
    return (block<0) ? 0 : track_find(-1);
  }

  for (i=0,notplayed=0 ; i<nfound ; i++)
   if (trks[found[i]].last_played==0) notplayed++;

  // We have some tracks that haven't been played, so limit to just those
  if (notplayed)
  { for (i=0,notplayed=0 ; i<nfound ; i++)
     if (trks[found[i]].last_played==0) found[notplayed++]=found[i];
    nfound=notplayed;
  }

  i=rand()%nfound;
//  printf("%d to do, entry %d is %d\n",nfound,i,found[i]);
  i=found[i];
  free(found);
  return i;
}
