/* 
 * Jukebox interface
 * -------------------
 * copyright (c) 1997-1999 by Roger Walker, All rights reserved.
 * See the file RNW_COPY
 *
 * Uses the routines in disp_com.c or disp_crt.c to provide a user interface
 *
 *  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
 *  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
 *  PURPOSE.
 */

#define BANNER_STR "Dave's Halloween party"

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#include <memory.h>
#ifdef __DJGPP__
#include <go32.h>
#endif
#include <assert.h>
#include "jukebox.h"
#include "disp_win.h"

extern trk_info_t *trks;
extern unsigned tracks_bya[], tracks_byt[];
extern int nsorted;

// Work in progress !
bool home_mode;


extern play_list_t toplay[];
extern int ntoplay;
extern bool really_exit;
extern window_t *pCurrentWin;

typedef enum { mode_playing, mode_special, mode_special2, mode_artist, mode_art_trk, mode_track, mode_trk_art, mode_reset } mode_t;



static void generate_lists(void);
static WORD current_time(void);
static void generate_entry(char *dest,int win,int pos);
static void cursor_undraw(void);
static void cursor_draw(int win,int width,int pos,char *cursor);
static void print_entry(int pos,char *str);
static void status_update_window(int old_pos,int new_pos,int win_id);
static void status_update_current(int old_pos);
static void status_update_mode(char *help,char *mode);
static BYTE get_buttons(void);
static int password_checker(BYTE chgot);
static void update_special_status(void);
static void set_volumes(void);
static void change_mode(mode_t new_mode);


static mode_t current_mode;
static int debug_flags;
static int last_selected=0;
static int current_win,current_pos,current_step;
static int *current_sel,current_max;
static int *art_index,*art_list,nart;
static int *trk_index,*trk_list,ntrk;
static char *modes[mode_reset-mode_playing]=
        { "Play list", "Special Functions", "Special 2","Sorted by Artist", "Tracks by an artist", "Sorted by track name", "Track's artists" };
static char *help_str[mode_reset-mode_playing]=
{	"CANCEL=Exit   FAST+SEL=Delete   SEL=Play next",
    "CANCEL=Exit, U=Vol, D=Max, SEL=Just screen, FAST=Just play",
    "CANCEL=Exit, UP = Home mode, FAST + SEL=Exit program",
	"CANCEL=Select by Track          SEL=Select artist",
	"CANCEL=Choose another artist    SEL=Add track to playlist",
	"CANCEL=Select by artist         SEL=Select Track",
	"CANCEL=Choose another track     SEL=Add track to playlist" };

static int banner_pos;
static WORD banner_time;
static char *banner_string=
"Welcome to " BANNER_STR "        "
"Welcome to " BANNER_STR "        "
"Welcome to " BANNER_STR "        "
"      See  http://www.fareham.org/newjuke.htm for details about this box. ";
//"This unit was designed, programmed, and built by Roger     ""See the homepage at http://www.fareham.org/juke.htm        ""or e-mail   roger@fareham.org";

static int play_limit;

static WORD last_screen_change=0;

static char last_cur_data[1024];
static int last_cur_win,last_cur_x,last_cur_y,last_cur_wid=0;


static void generate_lists(void)
{ int i,j,lastPos;
  trk_info_t *pTrks;
  char *pLast;

  // Generate artist index, used to find their tracks in the tracks_bya[] list
  art_index=malloc(nsorted*sizeof(int)); assert(art_index!=NULL);
  nart=0;
  pLast=NULL;
  for (i=0 ; i<nsorted ; i++)
  { pTrks= &trks[tracks_bya[i]];
    // Check if same artist
	if (pLast!=NULL && (pTrks->artist==pLast || stricmp(pTrks->artist,pLast)==0)) continue;

	// New artist so add it
	art_index[nart++]=i;
	pLast=pTrks->artist;
  }
  // Add end entry which marks end of list
  art_index[nart]=nsorted;

  // Generate list of artists to display
  art_list=malloc(nart*sizeof(int)); assert(art_list!=NULL);
  for (i=0 ; i<nart ; i++)
  { art_list[i]=tracks_bya[art_index[i]];
  }

  // Generate track index
  trk_index=malloc(nsorted*sizeof(int)); assert(trk_index!=NULL);
  ntrk=0;
  pLast=NULL;
  for (i=0 ; i<nsorted ; i++)
  { pTrks= &trks[tracks_byt[i]];
    // Check if same track
	if (pLast!=NULL && (pTrks->title==pLast || stricmp(pTrks->title,pLast)==0)) continue;

	// New track so add it
	trk_index[ntrk++]=i;
	pLast=pTrks->title;
  }
  // Add end entry which marks end of list
  trk_index[ntrk]=nsorted;

  // Generate list of tracks to display
  trk_list=malloc(ntrk*sizeof(int)); assert(trk_list!=NULL);
  for (i=0 ; i<ntrk ; i++)
  { trk_list[i]=tracks_byt[trk_index[i]];
  }
}


/* Maximum of 60 second, polling at least every .20 */
#define TIME_MAX		6000

/* Get time in 1/100th second units. Range 0-5999 (1 minute) */
static WORD current_time(void)
{ struct dostime_t time;
  _dos_gettime(&time);
  return (time.second*100L + time.hsecond)%TIME_MAX;
}


static void generate_entry(char *dest,int win,int pos)
{ trk_info_t *pTrk;
  dest[0]='\0';
  if (win==WIN_STATUS)
  { if (pos<ntoplay)
	{ pTrk=&trks[toplay[pos].trk_index];
	  strcpy(dest,pTrk->artist);
	  strcat(dest," - ");
	  strcat(dest,pTrk->title);
	}
  } else
  if (pos<current_max)
  { pTrk=&trks[current_sel[pos]];
	switch(current_mode)
	{ case mode_artist:
	  case mode_trk_art:
		strcpy(dest,pTrk->artist);
		break;
	  case mode_art_trk:
	  case mode_track:
		strcpy(dest,pTrk->title);
		break;
	}
  }
}


static void cursor_undraw(void)
{ if (last_cur_wid==0) return;
  window_select(last_cur_win);
  window_move_xy(last_cur_x,last_cur_y);
  window_hilite(0);
  window_print_n(" ",1);
  window_print_n(last_cur_data,last_cur_wid);
  last_cur_wid=0;
}


static void cursor_draw(int win,int width,int pos,char *cursor)
{ last_cur_win=win;
  last_cur_x=((pos/pCurrentWin->height)%pCurrentWin->cols)*width;
  last_cur_y=pos%pCurrentWin->height;
  last_cur_wid=width-1;

  generate_entry(last_cur_data,win,pos);

  window_move_xy(last_cur_x,last_cur_y);
  window_print_n(cursor,1);
  window_hilite(1);
  window_print_n(last_cur_data,last_cur_wid);
  window_hilite(0);
}


static void print_entry(int pos,char *str)
{ int wid;

  wid=(pCurrentWin->width/pCurrentWin->cols);
  window_move_xy(((pos/pCurrentWin->height)%pCurrentWin->cols)*wid,pos%pCurrentWin->height);
  window_print_n(str,1);
}

// Return base 62 value
static char seqnum(int val)
{ if (val==99999) return '>';
  val%=52;
  if  (val<26) return ('A'+val);
  val-=26;
  if  (val<26) return ('a'+val);
  val-=26;
  return '0'+val;
}

static void status_update_window(int old_pos,int new_pos,int win_id)
{ int actual_pos,pos,wid,per_page;
  char new_cursor[2];
  
  window_select(win_id);
  wid=(pCurrentWin->width/pCurrentWin->cols);
  per_page=pCurrentWin->height*pCurrentWin->cols;

  // Set default cursor as >
  strcpy(new_cursor,">");
  
  // Check if this track has been played
  if (win_id==WIN_POPUP)
  { int i;
	for (i=0 ; i<ntoplay ; i++)
	 if (toplay[i].trk_index==current_sel[new_pos])
	 { strcpy(new_cursor,"*");
	   break;
	 }
  } else
  if (win_id==WIN_STATUS)
  { new_cursor[0]=seqnum(toplay[new_pos].sequence);
  }

  if (current_win==win_id && (old_pos/per_page)==(new_pos/per_page))
  { // Same page so just redraw the cursor
	if (old_pos>=0)
	{ cursor_undraw(); // print_entry(old_pos," ");
	}

	cursor_draw(win_id,wid,new_pos,new_cursor);
	//print_entry(new_pos,new_cursor);
	return;
  }

  last_screen_change=current_time();

  for (pos=0,actual_pos=(new_pos/per_page)*per_page ; pos<per_page ; pos++,actual_pos++)
	if (current_win==win_id && actual_pos==new_pos)
	{ cursor_draw(win_id,wid,actual_pos,new_cursor);
	} else
	{ char buf[256];
	  generate_entry(buf,win_id,actual_pos);
	  window_move_xy((pos/pCurrentWin->height)*wid,pos%pCurrentWin->height);
	  window_print_n(" ",1);
	  window_print_n(buf,wid-1);
	}
}


static void status_update_current(int old_pos)
{ if (current_pos<0 || current_max==0) current_pos=0; else
  if (current_pos>=current_max) current_pos=current_max-1;
  status_update_window(old_pos,current_pos,current_win);
}

static void status_update_mode(char *help,char *mode)
{ window_select(WIN_ERR);
  window_move_xy(0,0);
  window_print_n(help,60);
  window_print_n(mode,20);
}


/* 
 *
 * Uses the parallel port to connect to the physical hardware
 * Data outputs 0-7 are connected to the buttons via diodes
 * The return is connected to ERROR
 *
 */

/* LPT port, port addresses and status bits */
static WORD lpt_port;
#define PORT_DTR		(lpt_port+0)

#define PORT_STR		(lpt_port+1)
#define STR_ERR			0x08
#define STR_SLCT		0x10
#define STR_PE			0x20
#define STR_ACK			0x40
#define STR_BUSY		0x80
#define STR_INPUT_MASK	(STR_ERR|STR_SLCT|STR_PE|STR_ACK|STR_BUSY)
#define STR_INVERT		(STR_ERR|STR_SLCT|STR_PE|STR_ACK)

#define PORT_CTR (lpt_port+2)
#define CTR_STOBE		0x01
#define CTR_AFD			0x02
#define CTR_INIT		0x04
#define CTR_SELECT		0x08
#define CTR_CONFIGURE	CTR_STOBE|CTR_AFD|CTR_INIT|CTR_SELECT


#define BUTTON_CANCEL	(1<<0)
#define BUTTON_DOWN		(1<<1)
#define BUTTON_UP		(1<<2)
#define BUTTON_SEL		(1<<3)
#define BUTTON_FAST		(1<<4)

static int next_sequence=32767;

static BYTE get_buttons(void)
{ BYTE down,pcfake,bitval;
  int i;
  union REGS rio;

  down=(lpt_port==0) ? 0 : ((inp(PORT_STR)&STR_INPUT_MASK)^STR_INVERT)>>3;

  // Remove illegal key combinations early
  if ((down&(BUTTON_DOWN|BUTTON_UP))!=0 && (down&(BUTTON_SEL|BUTTON_CANCEL))!=0)
  { // UP or DOWN with SEL or CANCEL
	down&= ~(BUTTON_DOWN|BUTTON_UP|BUTTON_SEL|BUTTON_CANCEL);
  } else
  if ((down&(BUTTON_DOWN|BUTTON_UP))==(BUTTON_DOWN|BUTTON_UP))
  { // UP and DOWN
	down&= ~(BUTTON_DOWN|BUTTON_UP);
  } else
  if ((down&(BUTTON_SEL|BUTTON_CANCEL))==(BUTTON_SEL|BUTTON_CANCEL))
  { // SEL and CANCEL
	down&= ~(BUTTON_SEL|BUTTON_CANCEL);
  }
  
  rio.h.ah=2; int86(0x16,&rio,&rio); 
  pcfake=rio.h.al&0xF;
  // Read PC keyboard
  // Bit 0 = Right shift
  // Bit 1 = Left shift
  // Bit 2 = Either CTRL
  // Bit 3 = Either ALT
  //

  // Check for abort
  if (really_exit!=0)
  { // We are already trying to abort, so don't cofuse things with more key hits
	return 0;
  } else
  if ((pcfake&7)==7)
  {	abort_playing();
	really_exit=1;
  }

  // Emulate the real keys
  if ((pcfake&10)==10 && current_mode!=mode_playing)	// Left shift and ALT == FAST + CANCEL (only if mode!=playmode)
  { down|=BUTTON_CANCEL;
    pcfake&= ~8;
  }
  if ((pcfake&5)==5)	// Right shift & control == CANCEL
  { down|=BUTTON_CANCEL;
	pcfake&= ~5;
  }
  if ((pcfake&1)==1)	// Right shift = UP
  { down|=BUTTON_UP;
    pcfake&= ~1;
  }
  if ((pcfake&4)==4)	// Control = DOWN
  { down|=BUTTON_DOWN;
    pcfake&= ~4;
  }
  if ((pcfake&2)==2)	// Left shift = FAST
  { down|=BUTTON_FAST;
    pcfake&= ~2;
  }
  if ((pcfake&8)==8)	// Alt = Select
  { down|=BUTTON_SEL;
    pcfake&= ~8;
  }

  return down;
}


#define DOWN2   0xCB // ''
#define BOTH2   0xCE // ''
#define UP2     0xCA // ''
#define DOWN1   0xD1 // ''
#define BOTH1   0xD8 // ''
#define UP1     0xCF // ''
#define HORIZ2	0xCD

#define NDIGITS 4

static int password_active;

static int password_checker(BYTE chgot)
{ static int ngot;
  static int toget;
  int x,y,ok;
  static BYTE codes[5]={ DOWN1,UP1,DOWN2,UP2,HORIZ2 };

  // Find position to draw character
  {	extern window_t windows[];
	window_t *pWin;
	pWin= &windows[WIN_MAIN];
	x=pWin->left+pWin->width-1;
	y=pWin->top+pWin->height;
  }

  if (chgot==0)
  { // Initialise values
	ngot=0;
	ok=1;
  } else
  { // Check for only the UP XOR the DOWN buttons
	ok=((chgot&BUTTON_UP)==0 ^ (chgot&BUTTON_DOWN)==0);
	// Check for UP or DOWN correct
	if (ok) ok=((toget&1)==0 ^ (chgot&BUTTON_UP)!=0);
	// Check for fast
	if (ok) ok=((toget&2)==0 ^ (chgot&BUTTON_FAST)!=0);
	ngot++;
  }

  toget=rand()%4;
  if (!ok || ngot==NDIGITS)
  { for (toget=0 ; toget<ngot ; toget++) display_char(toget,y+1,' ');
	toget=4;
  }
  window_hilite(2);
  display_char(x,y,codes[toget]);
  window_hilite(0);

  if (ok)
  { if (ngot==NDIGITS)
	{ change_mode(mode_playing);
	  password_active=0;	  
	} else
	{ display_char(ngot,y+1,'*');
	}
	return 0;
  } else
  { password_active=0;
    return 1;
  }
}


// Can't do volume at init, because the sound card isn't sorted out yet !
static int req_volume,last_volume= -1;

// Display the volume control, and if in mode_special the max number of tracks
static void update_special_status(void)
{ extern window_t windows[];
  window_t *pWin;
  char buf[8];
  int i;

  // Only show max track if in special mode
  if (current_mode==mode_special)
  { sprintf(buf,"%3u",play_limit);
    window_select(WIN_ERR);
    window_move_xy(77,0);
    window_print_n(buf,3);
  }

  // Find position to draw character
  pWin= &windows[WIN_MAIN];
  window_hilite(2);
  // Don't bother undrawing then re-drawing
  for (i=0 ; i<=VOL_MAXVAL ; i++)
	  display_char(i,pWin->top-1,(i==req_volume) ? UP1 : HORIZ2);
  window_hilite(0);
}


// Try to set volume, if it works set them all
static void set_volumes(void)
{ if (req_volume==last_volume) return;

  if (volume_write(VOL_MASTER,   req_volume*VOL_SCALE))
  {	last_volume=req_volume;
    volume_write(VOL_WAVE, (VOL_MAXVAL*9/10)*VOL_SCALE);

	/* Mute (unused) line and CD in */
	volume_write(VOL_LINE,0);
	volume_write(VOL_CD,0);

	/* Set tones to midway */
	volume_write(TONE_BASE,(TONE_MAXVAL*TONE_SCALE/2));
	volume_write(TONE_TREBLE,(TONE_MAXVAL*TONE_SCALE/2));
	update_special_status();
  }
}


static void change_mode(mode_t new_mode)
{ int i,j;
  trk_info_t *pLastSel,*pTrk;
  static int last_current;
  static int playing_size;

  // Undraw the current cursor
  if (current_mode==mode_playing ||
	  current_mode==mode_artist ||
	  current_mode==mode_track)
  {	cursor_undraw();
	//window_select(current_win); print_entry(current_pos," ");
  }

  if (current_mode==mode_playing)
  { // Check if the cursor is away from the home screen
	if (current_pos>=playing_size)
	{ status_update_window(-9999,0,WIN_STATUS);
	  window_move_xy(0,0); 
	  window_print_n(" ",1);	
	}
  }

  // Pretend that mode_special is really mode_playing
  if (current_mode==mode_playing && new_mode==mode_special)
  { current_mode=new_mode;
    status_update_mode(help_str[current_mode],modes[current_mode]);
	update_special_status();
    return;
  } else
  if (current_mode==mode_special)
  { current_mode=mode_playing;
  }

  pLastSel= &trks[last_selected];

  switch(new_mode)
  { case mode_playing:
		current_win=WIN_STATUS;
		current_max=ntoplay;
		current_sel=NULL;
		current_pos= -1;
		// Try to find the last selected track
		for (i=0 ; i<ntoplay ; i++)
		if (toplay[i].trk_index==last_selected)
		{ current_pos=i;
		  break;
		}
		// If not found, try to find something by the same artist
		if (current_pos<0)
		for (i=0 ; i<ntoplay ; i++)
		{ pTrk= &trks[toplay[i].trk_index];
		  if (pTrk->artist==pLastSel->artist || (stricmp(pTrk->artist,pLastSel->artist)==0))
		  { current_pos=i;
		    break;
		  }
		}
		break;

	case mode_artist:
		current_win=WIN_MAIN;
		current_max= nart;
		current_sel= art_list;
		if (current_mode==mode_art_trk)
		{ current_pos=last_current;
		  break;
		}
		current_pos= -1;
		for (i=0 ; i<nsorted ; i++)
		if (tracks_bya[i]==(unsigned)last_selected)
		{ // Found track
		  for (j=0 ; j<nart ; j++)
			if (art_index[j]<=i && i<art_index[j+1])
			{ current_pos=j;
			  break;
			}
		  break;		  
		}
		if (current_pos<0) current_pos= current_max/2;
		break;

	case mode_track:
		current_win=WIN_MAIN;
		current_max= ntrk;
		current_sel= trk_list;
		if (current_mode==mode_trk_art)
		{ current_pos=last_current;
		  break;
		}
		current_pos= -1;
		for (i=0 ; i<nsorted ; i++)
		if (tracks_byt[i]==(unsigned)last_selected)
		{ // Found track
		  for (j=0 ; j<ntrk ; j++)
			if (trk_index[j]<=i && i<trk_index[j+1])
			{ current_pos=j;
			  break;
			}
		  break;		  
		}
		if (current_pos<0) current_pos= current_max/2;
		break;

	case mode_art_trk:
		assert(current_mode==mode_artist);
		last_current=current_pos;
		current_win=WIN_POPUP; window_select(current_win); window_border();
		i=art_index[current_pos];				// Lookup the position
		current_max=art_index[current_pos+1]-i;	// Find distance to next position
		current_sel= &tracks_bya[i];			// Find start in tracks_bya list
		current_pos=current_max/2;				// Set default start pos
		for (i=0 ; i<current_max ; i++)
		 if (last_selected==current_sel[i])
		 { current_pos=i;
		   break;
		 }
		break;

	case mode_trk_art:
		assert(current_mode==mode_track);
		last_current=current_pos;
		current_win=WIN_POPUP; window_select(current_win); window_border();
		i=trk_index[current_pos];				// Lookup the position
		current_max=trk_index[current_pos+1]-i;	// Find distance to next position
		current_sel= &tracks_byt[i];			// Find start in tracks_bya list
		current_pos=current_max/2;				// Set default start pos
		for (i=0 ; i<current_max ; i++)
		 if (last_selected==current_sel[i])
		 { current_pos=i;
		   break;
		 }		
		break;
  }  

  current_mode=new_mode;
  status_update_current(-9999);
  current_step=pCurrentWin->height;
  if (current_step<10) current_step=pCurrentWin->height*pCurrentWin->cols;

  if ((current_mode==mode_art_trk || current_mode==mode_trk_art) &&
		ntoplay>=play_limit)
  { char *msg=" TOO MANY TRACKS: Wait for one to play ! ";
    // We are in an add track mode, but there is no space to add a track !
	window_move_xy((pCurrentWin->width-strlen(msg))/2,-1);
	window_hilite(1);
	window_print(msg);
	window_hilite(0);
  }
  
  // Remember the mode_playing window size
  if (current_mode==mode_playing)
  { playing_size=pCurrentWin->height*pCurrentWin->cols;
  }
  status_update_mode(help_str[current_mode],modes[current_mode]);
}

// ************************************************************************************
// ************************************************************************************
// ************************************************************************************
// EXPORTED FUNCTIONS / Variables

void display_playing_status(int secs,int busy)
{ char buf[128];
  window_select(WIN_ERR);
  window_move_xy(70,1);
  sprintf(buf,"%2u:%02u ",secs/60,secs%60);
  window_print_n(buf,6);

  sprintf(buf,"%4u",busy);
  window_print_n(buf,4);
}


void status_update_tracklist(void)
{ static int ndone;
  if (current_mode==mode_playing)
  { current_max=ntoplay;
    status_update_current(-9999);
    if (current_max==0)
	{ window_move_xy(0,0);
	  window_print_n("*",1);
	}
  } else
  { status_update_window(-9999,0,WIN_STATUS);
	window_move_xy(0,0); window_print_n(" ",1);
	if (debug_flags)
	{ char buf[10];
	  window_move_xy(74,0);
	  sprintf(buf,"%d",ndone++);
	  window_print_n(buf,6);	  
	}
  }
}

// The DMA output buffer is full (or we're not playing anything) so sort out the display
void run_idle_task(void)
{ status_update();
  window_select(WIN_ERR); window_move_xy(0,0);
}

char *track_label(unsigned trk)
{ static char retbuf[8];
  sprintf(retbuf,"%5u",trk);
  return retbuf;
}

void status_reset(void)
{ char *env;
  WORD start;

  // Find the parallel port and set the default outputs
#ifdef __DJGPP__
   lpt_port=_farpeekw(_go32_conventional_mem_selector(),0x408);
#else
  WORD far *fp;
  /* Find first LPT port and configure it with NO IRQs and required I/Os */
  *((DWORD *)(&fp))=0x00400008; /* Point to the LPT port BIOS data area */
  lpt_port= *fp;
#endif
  if (lpt_port==0) return;

  // Set outputs high to start with
  outp(PORT_DTR,0);			// Set all data lines low
  outp(PORT_STR,0);			// Reset status lines
  outp(PORT_CTR,CTR_INIT);	// Set all control lines high

  /* Set volumes to 90% */
  req_volume=VOL_MAXVAL*9/10;

  // Sort out the main screen
  display_init();
  window_select(WIN_MAIN);
  window_border();
  window_select(WIN_ERR); 
  window_move_xy(0,0);
  
  current_mode=mode_reset;
  env=getenv("JUKEBOX");
  debug_flags=(env==NULL) ? 0 : strtol(env,NULL,0);

  while (((inp(PORT_STR)&STR_INPUT_MASK)^STR_INVERT)!=0)
  { // We have a button held down !
	window_move_xy(0,0);
	window_print("Release the key on the keypad, or press a key on the real keyboard");
	if (kbhit())
	{ if (getch()==0) getch();
	  lpt_port=0;
	  debug_flags|=1;
	  break;
	}	
  }
  window_move_xy(0,0); window_clear_n(80);

  if (debug_flags)
  { window_move_xy(0,1);
    window_print("RShift+CTRL=Cancel, Right Shift=UP, CTRL=Down, LShift=Fast, Alt=SEL");
  }

  generate_lists();
  status_update_mode("Press a button to start",debug_flags ? "Debug mode" : "Init done");
  window_select(WIN_MAIN);

  banner_time=current_time();
  banner_pos=0;
  last_cur_wid=0;

  play_limit=20; // 20 tracks is about an hour
}


void status_refresh(void)
{ status_update_tracklist();
}


// Scan keys and work out mode
void status_update(void)
{ BYTE buts,bitval;
  int step,last_pos,i;
  WORD time_now,this_down;
  static WORD time_down[4];

  // Get the current time to 1/100 sec
  time_now=current_time();

  // Caclulate time since last screen change
  i=(time_now-last_screen_change+TIME_MAX)%TIME_MAX;

  // If it was less than 250mS ago, then ignore
  if (i<25) return;
  // Force it to exactly 250mS ago, so that if it's ages to the next poll things are OK
  last_screen_change=(time_now+TIME_MAX-25)%TIME_MAX;

  // We can't be sure of when this happens so try it whenever it's wrong
  if (req_volume!=last_volume) set_volumes();

  // Read the Parallel port and the PC keyboard emulation
  buts=get_buttons();

  // Initial reset, waits for a key before changing mode
  if (buts!=0 && current_mode==mode_reset)
  { change_mode(mode_artist);
    // Mark all keys as down now, will cause inital keypress to be ignored
    for (i=0 ; i<4 ; i++) time_down[i]=time_now;
    return;
  }

  // Auto repeat the UP and DOWN buttons, debounce the SEL and CANCEL
  for (i=0,bitval=1 ; i<4 ; i++,bitval<<=1)
   if ((buts&bitval)==0)
   { // This button has been released so mark it as such
	 time_down[i]=TIME_MAX;
   } else
   { if (time_down[i]==TIME_MAX)
	 { // If wasn't already down use it and rememeber when first down
	   time_down[i]=time_now;
	   this_down=TIME_MAX;
	 } else
	 { // Else calculate how long it has actually been down
	   this_down=(time_now-time_down[i]+TIME_MAX)%TIME_MAX;
	 }

	 // We have a button down, and this_down=time down or TIME_MAX if just down

	 if (bitval==BUTTON_DOWN || bitval==BUTTON_UP)
	 { WORD inital=50,repeat=15;
	   if (this_down<inital)
	   { // Down a while, but not long enough to auto repeat so clear key
		 buts&= ~bitval;
	   } else
	   if (this_down!=TIME_MAX)
	   { // Not just down, but long enough to start auto repeat
		 // Fake time_down to cause auto repeat sooner at repeat time rather than inital time
		 time_down[i]=(time_now-inital+repeat+TIME_MAX)%TIME_MAX;
	   }
	 } else
	 if (bitval==BUTTON_SEL || bitval==BUTTON_CANCEL)
	 { WORD inital=5;
	   // In certain modes slow down the response
	   if ((bitval==BUTTON_CANCEL && (current_mode==mode_artist || current_mode==mode_track)) ||
		   (bitval==BUTTON_SEL && current_mode!=mode_artist && current_mode!=mode_track))
	   { inital=20;
	   }
	   if (this_down==TIME_MAX || this_down<inital)
	   { // Just down or not down for very long
	     buts&= ~bitval;
	   } else
	   if (time_down[i]>TIME_MAX)
	   { // Already signalled, so ignore
	     buts&= ~bitval;
	   } else
	   { // First time, mark as done
		 time_down[i]=TIME_MAX+1;
	   }
#if 0
	   else
	   if (this_down>(TIME_MAX/2))
	   { // Down for longer than TIME_MAX/2, so reset so that it never repeats
		 time_down[i]=(time_now-(TIME_MAX/4)+TIME_MAX)%TIME_MAX;		 
	     buts&= ~bitval;
	   }
#endif
	 }
   }

  //  if (debug_flags==0)
  { i=(time_now-banner_time+TIME_MAX)%TIME_MAX;
    if (i>=25)
    { banner_time=time_now;
      window_select(WIN_ERR);
      window_move_xy((banner_pos>=70) ? 0 : (70-banner_pos),1);
      window_print_n(banner_string+((banner_pos<70) ? 0 : banner_pos-70),(banner_pos<70) ? banner_pos : 70);
      banner_pos++;
      banner_pos%=(strlen(banner_string)+71);      
    }
  }

  // No buttons down (except maybe FAST) so don't need to check for change of status
  if ((buts&(BUTTON_UP|BUTTON_DOWN|BUTTON_SEL|BUTTON_CANCEL))==0) return;

  // Password protect some of the menus
  if (password_active)
  { if (password_checker(buts)==0) return;
    change_mode(mode_artist);
	return;
  }




  // Special mode, allows change of volume and max track limit
  if (current_mode==mode_special)
  { if (buts&BUTTON_CANCEL)
        { change_mode(((buts&BUTTON_FAST)==0) ? mode_artist : mode_special2);
	} else
	if (buts&BUTTON_SEL)
	{ int max_left;
	  // Delete all tracks except the screen playing (or one playing if FAST held down)
	  max_left=(buts&BUTTON_FAST) ? 1 : 8;
	  if (ntoplay>max_left)
	  { ntoplay=max_left+1;				// Force to 2 or 9 tracks (0..1 or 0..8)
	    track_delete_from(max_left);	// Delete track 1 or track 8
	  }
	} else
	if (buts&BUTTON_UP)
	{ if (buts&BUTTON_FAST)
	  { if (req_volume<VOL_MAXVAL) req_volume++;
	  } else
	  { if (req_volume>0) req_volume--;
	  }
	} else
	if (buts&BUTTON_DOWN)
	{ if (buts&BUTTON_FAST)
	  { if (play_limit==5) play_limit+=5; else
	    if (play_limit<MAX_PLAY) play_limit+=10;
	  } else
	  { if (play_limit==10) play_limit-=5; else
	    if (play_limit>10) play_limit-=10;
	  }
	  update_special_status();
	}

	if (ntoplay>play_limit)
	{ ntoplay=play_limit;
	  status_update_tracklist();
	}
	// In special mode don't allow anthing else to see the buttons
	return;
  } else
  if (current_mode==mode_special2)
  { if (buts&BUTTON_CANCEL)
	{ change_mode(((buts&BUTTON_FAST)==0) ? mode_artist : mode_playing);
	} else
	if (buts&BUTTON_SEL)
	{ int max_left;
          // Exit from the player software
          if (buts&BUTTON_FAST)
           {  // Check for abort
                  if (really_exit!=0)
                  { // We are already trying to abort, so don't cofuse things with more key hits
                       return;
                  } else
                  {     FILE *out;
                               out=fopen("c:\\scandisk.run","wt");
                               fputc('.',out);
                               fclose(out);
                               abort_playing();
                        really_exit=1;
                  }
           }
	} else
	if (buts&BUTTON_UP)
        { home_mode=1;
	} else
	if (buts&BUTTON_DOWN)
        { home_mode=0;
        }
	// In special mode don't allow anthing else to see the buttons
	return;
  }




  // These menus have track lists associated with them.
  step=(buts&BUTTON_FAST) ? current_step : 1;

  if (buts&BUTTON_SEL)
  {	switch(current_mode)
	{ case mode_playing:
		if (ntoplay==0)
		{ status_update_tracklist();
		} else
		if (buts&BUTTON_FAST) // FAST & SEL = Delete
		{ track_delete_from(current_pos);
		  status_update_tracklist();
		  last_selected=(ntoplay==0) ? 0 : toplay[current_pos].trk_index;
		} else 
		if (current_pos>0 && current_pos<ntoplay) // Just SEL = Move to next to play (Only sensible for tracks 2+)
		{ char new_cursor[2];
		  last_selected=toplay[current_pos].trk_index;
		  toplay[current_pos].sequence=(next_sequence--);
		  new_cursor[0]=seqnum(toplay[current_pos].sequence);
		  new_cursor[1]='\0';
		  window_select(current_win);
		  print_entry(current_pos,new_cursor);
		}		
		break;

	  case mode_artist:
		change_mode(mode_art_trk);
		break;

	  case mode_track:
		change_mode(mode_trk_art);
		break;

	  case mode_art_trk:
	  case mode_trk_art:
		// We are at the maximum limit, ignore it
		if (ntoplay>=play_limit) break;

		last_selected=current_sel[current_pos];
		track_add(last_selected);
		window_select(current_win);
		print_entry(current_pos,"*");

		status_update_tracklist();
		if (debug_flags==0)
		{ // If not debug mode, exit this menu after something is selected
		  change_mode((current_mode==mode_art_trk) ? mode_artist : mode_track);
		}
		break;
	}
	buts=0; // Ignore other keys after a mode change
  }

  if (buts&BUTTON_CANCEL)
  {	mode_t new_mode;
    if ((current_mode==mode_track || current_mode==mode_artist) && (buts&BUTTON_FAST)!=0)
	{ password_checker(0);
	  password_active=1;
	} else
	if (current_mode==mode_playing && (buts&BUTTON_FAST)!=0)
	{ change_mode(mode_special);
	} else
	{ new_mode=	(current_mode==mode_playing) ? mode_artist :
				(current_mode==mode_artist) ? mode_track :
				(current_mode==mode_track) ? mode_artist:
				(current_mode==mode_art_trk) ? mode_artist :
				(current_mode==mode_trk_art) ? mode_track : mode_artist;
	  if (new_mode!=current_mode) change_mode(new_mode);
	}

	buts=0; // Ignore other keys after a mode change
  }

  last_pos=current_pos;
  if (buts&BUTTON_UP)
  {	current_pos=(current_pos>=step) ? (current_pos-step) : 0;
  }

  if (buts&BUTTON_DOWN)
  {	current_pos+=step;
    if (current_pos>=current_max) current_pos=(current_max-1);
  }

  if (debug_flags)
  {	char buf[128];
	sprintf(buf,"%02X %d %3d %3d %3d",buts,current_mode,last_pos,current_pos,step);
	if (current_sel!=NULL)
	{ trk_info_t *pTrk;
	  pTrk= &trks[current_sel[current_pos]];
	  sprintf(buf+strlen(buf)," %8p %8p",pTrk->artist,pTrk->title);
	}
	window_select(WIN_ERR);
	window_move_xy(40,0);
	window_print_clr(buf);
  }

  if (last_pos!=current_pos)
  {	window_select(current_win);
	status_update_current(last_pos);
  }
}
