/* 
 * Jukebox interface
 * -------------------
 * copyright (c) 1997-1999 by Roger Walker, All rights reserved.
 * See the file RNW_COPY
 *
 * Uses the routines in disp_com.c or disp_crt.c to provide a user interface
 *
 *  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
 *  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
 *  PURPOSE.
 */
#include <conio.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include "jukebox.h"

extern trk_info_t *trks;
extern unsigned int nsorted;

extern play_list_t toplay[];
extern int ntoplay;
extern int tracks_bya[], tracks_byt[], local_cd[];
extern int really_exit;
extern int display_height;


static WORD secs_remaining;
static enum { s_reset,s_sel_a,s_sel_t,s_cue } status;
static int refresh,*list_sel,list_pos,list_last,list_nitems;
static int last_ntoplay;
static int *play_list=NULL;

static int last_trk_disp;

#define LINE_LEN 79
#define ROW_KEYS1 0
#define ROW_KEYS2 1
#define ROW_ERROR 2
#define ROW_CUE   3
#define ROW_PLAY  4
#define ROW_LIST  6
#define LIST_DISP (display_height-ROW_LIST)

static int display_str_n(char *str,int maxlen)
{ char tmp[1024]; int len;
  if (maxlen<=0) return 0;
  len=strlen(str);
  if (len<=maxlen)
   { display_str(str); return len;
   }
  strcpy(tmp,str); tmp[maxlen]='\0';
  display_str(tmp);
  return maxlen;
}

static int *regen_playing(void)
{ int i;
  if (play_list!=NULL)
  { free(play_list); play_list=NULL;
  }
  play_list=malloc(ntoplay*sizeof(int));
  for (i=0 ; i<ntoplay ; i++)
	  play_list[i]=toplay[i].trk_index;
  return play_list;
}

static int row_data[50];

static void display_entry(int row,char *prefix,int trk_num)
{ int nch;
  trk_info_t *tp;

  display_row(row);
  display_str(prefix);

  if (row_data[row]==trk_num) return;

  nch=strlen(prefix);
  tp=trks+trk_num;
  nch+=display_str_n(" ",LINE_LEN-nch);
  nch+=display_str_n(tp->artist,LINE_LEN-nch);
  nch+=display_str_n(" - ",LINE_LEN-nch);
  nch+=display_str_n(tp->title,LINE_LEN-nch);
  display_clr_eol();
  row_data[row]=trk_num;
}

static void display_cue_status(void)
{ char buf[64];
  sprintf(buf,"Total of %d track%s cued up",ntoplay,(ntoplay!=1) ? "s" :"");
  display_row(ROW_CUE);
  display_str(buf);
  display_clr_eol();
}

void display_playing_status(int secs,int busy)
{ char buf[64];
  if (ntoplay>0)
   { sprintf(buf,"%3d %3d:%02d",busy,secs/60,secs%60);
	 if (last_trk_disp!=toplay[0].trk_index)
	 { display_entry(ROW_PLAY,buf,toplay[0].trk_index);
	   last_trk_disp=toplay[0].trk_index;
     } else
	 { display_row(ROW_PLAY);
	   display_str(buf);
	 }
   } else
  if (last_trk_disp!= -1) /* Only display this once */
   { display_row(ROW_PLAY);
     display_str("Nothing playing");
	 display_clr_eol();
	 last_trk_disp= -1;
   }
}

static void display_list(int *list,int pos,int max)
{ int i,cursor;
  if (pos<LIST_DISP/2 || max<=LIST_DISP)
   { /* Too near top, or not enough entries */
     cursor=pos; pos=0;
   } else
  if ((pos+LIST_DISP/2)>max)
   { /* Too near bottom */
     cursor=pos;
     pos=max-LIST_DISP;
     cursor-=pos;
   } else
   { cursor=LIST_DISP/2; pos-=cursor;
   }
  cursor+=ROW_LIST;

  if (display_cancel())
  { // Cancel actually cleared some data, so might need to redraw lines
	memset(row_data,-1,sizeof(row_data));
  }
  for (i=ROW_LIST ; i<=ROW_LIST+LIST_DISP ; i++,pos++)
   if (pos<0 || pos>max)
	{ if (row_data[i]!= -1)
	  { display_row(i); 
	    display_clr_eol(); 
        row_data[i]= -1;
	  }
    } else
    { display_entry(i,(i==cursor) ? "> " : "  ",list[pos]);
    }
}


static void actual_display_update(void)
{ int i;

  /* Limit list point in case we are on cue list and things have changed */
  if (list_nitems>0)
   { if (list_pos>list_nitems-1) list_pos=list_nitems-1; else
     if (list_pos<0) list_pos=0;
   }

  if (refresh)
  { display_clear();
    memset(row_data,-1,sizeof(row_data));
    display_row(ROW_KEYS1);
    display_str("Display: F1=By Artist, F2=By Track, F4=Current selections.");
    
	last_trk_disp!= -2;
    display_row(ROW_KEYS2);
    switch (status)
    { case s_sel_a: case s_sel_t:
       display_str("ENTER = Add track");
       break;
      case s_cue:
       display_str("DELETE = remove track");
       break;
    }

    display_cue_status();
    last_trk_disp= -2;
	display_flush();
    refresh=0;
  }

  if (list_last!=list_pos)
  { // If showing cued tracks and number has changed update
    display_list(list_sel,list_pos,list_nitems-1);
    list_last=list_pos;
  }

  display_row(ROW_ERROR);
}



// ************************************************************************************
// ************************************************************************************
// ************************************************************************************
// EXPORTED FUNCTIONS / Variables

void status_update_tracklist(void)
{
  display_cue_status();
  if (status==s_cue)
  { refresh=1; list_last=list_pos+1; list_nitems=ntoplay;
    list_sel=regen_playing(); 
    actual_display_update();
  }
  display_row(ROW_ERROR);
}

// The DMA output buffer is full (or we're not playing anything) so sort out the display
void run_idle_task(void)
{ status_update();
  actual_display_update();
  display_row(ROW_ERROR);
}

char *track_label(int trk)
{ static char retbuf[8];
  sprintf(retbuf,"%d",trk);
  return retbuf;
}

void status_reset(void)
{ 

  memset(row_data,-1,sizeof(row_data));

  last_ntoplay= -1;
  status=s_reset;
  status_update();
  refresh=1;
  actual_display_update();
}

void status_refresh(void)
{ actual_display_update();
}

// Scan keys and work out mode
void status_update(void)
{ int chgot,i;

  while (1)
  {
  /* If reset then fake a selection key */
  if (status==s_reset)
  { chgot=KEY_F1;
  } else
  { chgot=display_getkey();
    // No key, so exit
    if (chgot==0) break;
  }

#if 0
  if (chgot>=KEY_F5 && chgot<=KEY_F10) /* Process volume + key hits */
  { int vol,val,scale,maxval;
    vol=(chgot==KEY_F5) ? VOL_MASTER :
        (chgot==KEY_F6) ? VOL_WAVE :
        (chgot==KEY_F7) ? VOL_CD :
        (chgot==KEY_F8) ? VOL_LINE :
        (chgot==KEY_F9) ? TONE_BASE : TONE_TREBLE;
    if (chgot==KEY_F9 || chgot==KEY_F10)
    { maxval=TONE_MAXVAL; scale=TONE_SCALE;
    } else
    { maxval=VOL_MAXVAL; scale=VOL_SCALE;
    }
    val=volume_read(vol)/scale;
    if (val<maxval)
	{ volume_write(vol,(val+1)*scale);
      display_volume_status();
	}
	return;
  }


  if (chgot>=KEY_SF5 && chgot<=KEY_SF10) /* Process volume - key hits */
  { int vol,val,scale;
    vol=(chgot==KEY_SF5) ? VOL_MASTER :
        (chgot==KEY_SF6) ? VOL_WAVE :
        (chgot==KEY_SF7) ? VOL_CD :
        (chgot==KEY_SF8) ? VOL_LINE :
        (chgot==KEY_SF9) ? TONE_BASE : TONE_TREBLE;
    scale=(chgot==KEY_SF9 || chgot==KEY_SF10) ? TONE_SCALE : VOL_SCALE;
    val=volume_read(vol)/scale;
    if (val>0)
	{ volume_write(vol,(val-1)*scale);
      display_volume_status();
	}
	return;
  }
#endif

  switch (chgot)     /* Process other key hits */
  { case KEY_F12:
     abort_playing();
     break;

    case KEY_ESC:
     if (ntoplay!=0) break;
     /* If we have nothing playing then treat ESC as ABORT */
    case KEY_CF12:
     abort_playing();
     really_exit=1;
     break;

    case KEY_F1:
	 if (status==s_sel_a) break;
     /* If we were looking by track then find same track in this list */
     if (status==s_sel_t || status==s_cue)
     { list_last=list_sel[list_pos];
       list_pos=0;
       while (list_pos<nsorted-1 && tracks_bya[list_pos]!=list_last)
          list_pos++;
	 } else
	 { list_pos=0;
	 }
     refresh=1; list_last=list_pos+1; 
	 list_nitems=nsorted; list_sel=tracks_bya; 
	 status=s_sel_a;
	 break;

    case KEY_F2:
	 if (status==s_sel_t) break;
     /* If we were looking by artist then find same track in this list */
     if (status==s_sel_a || status==s_cue)
	 { list_last=list_sel[list_pos];
       list_pos=0;
       while (list_pos<nsorted-1 && tracks_byt[list_pos]!=list_last)
          list_pos++;
	 } else
	 { list_pos=0;
	 }
     refresh=1; list_last=list_pos+1;
	 list_nitems=nsorted; list_sel=tracks_byt; 
	 status=s_sel_t;
	 break;


    case KEY_F4:
     list_pos=0; 
     refresh=1; list_last=list_pos+1; list_nitems=ntoplay;
     list_sel=regen_playing(); status=s_cue;
	 break;

    /* List scrolling keys */
    case KEY_HOME:
     list_pos=0;
     break;

    case KEY_END:
     if (list_nitems>0) list_pos=list_nitems-1;
     break;

    case KEY_UP:
     if (list_pos>0) list_pos--;
     break;

    case KEY_DOWN:
     if (list_pos<list_nitems-1) list_pos++;
     break;

    case KEY_PGUP:
     if (list_pos>=LIST_DISP) list_pos-=LIST_DISP;
     else list_pos=0;
     break;

    case KEY_PGDN:
     if (list_pos<((list_nitems-1)-LIST_DISP)) list_pos+=LIST_DISP;
     else list_pos=list_nitems-1;
     break;

    case KEY_DEL:
     if (status!=s_cue) break;
	 track_delete_from(list_pos);
	 list_sel=regen_playing();
	 list_nitems=ntoplay;
     break;

    case KEY_RET:
     if (status==s_cue) break;
//         if (ntoplay>=MAX_TABLE)
//          { fputs("\a",stdout); break;
//          }
     // Check if it's already in the list
     for (i=0 ; i<ntoplay ; i++)
          if (toplay[i].trk_index==list_sel[list_pos]) break;

     // If not then add it
     if (i>=ntoplay)
       track_add(list_sel[list_pos]);
     break;

	default:
      // If bad key ignore  
	  if (chgot>=256 || !isalnum(chgot)) break;

	  // If bad mode, ignore
	  if (status==s_cue) break;

      // Ignore case
      chgot=toupper(chgot);
	  
	  // Jump to first entry which starts with selected letter or number
	  list_pos=0;
      for (i=0 ; i<list_nitems ; i++)
      { trk_info_t *tp;
        char *cp;
        tp=trks+list_sel[i];
        cp=(status==s_sel_a) ? tp->artist : tp->title;
        if (toupper(*cp)>=chgot) 
        { list_pos=i;
		  break;
		}		
      }
      break;
  }
  }
}
