/* 
 * Jukebox constants
 * -------------------------------------
 * copyright (c) 1997-1999 by Roger Walker, All rights reserved.
 * See the file RNW_COPY
 *
 */

#ifndef __JUKEBOX_H
#define __JUKEBOX_H

/* Maximum number of discs loaded, and max number of tracks cued up */
#define MAX_PLAY 100
#define MAX_LOAD 4096

#ifdef __DJGPP__
#define _cdecl
#define ltoa itoa

#ifndef max
#define max(a,b) (((a) > (b)) ? (a) : (b))
#endif

#ifndef min
#define min(a,b) (((a) < (b)) ? (a) : (b))
#endif

#endif




/* Sound blaster mixer controls */
#define VOL_MASTER      0x30
#define VOL_WAVE        0x32
#define VOL_CD          0x36
#define VOL_LINE        0x38
#define VOL_SCALE       8
#define VOL_MAXVAL      31

#define TONE_BASE       0x44
#define TONE_TREBLE     0x46
#define TONE_SCALE      16
#define TONE_MAXVAL     15
         

#define KEY_BS		8
#define KEY_CBS		127
#define KEY_ABS		256+14

#define KEY_RET		13
#define KEY_CRET	10

#define KEY_ESC		27

#define TAB			9
#define STAB		256+15
#define CTAB		256+148

#define KEY_F1		256+59
#define KEY_F2		256+60
#define KEY_F3		256+61
#define KEY_F4		256+62
#define KEY_F5		256+63
#define KEY_F6		256+64
#define KEY_F7		256+65
#define KEY_F8		256+66
#define KEY_F9		256+67
#define KEY_F10		256+68
#define KEY_F11		256+133
#define KEY_F12		256+134
#define KEY_SF1		256+84
#define KEY_SF2		256+85
#define KEY_SF3		256+86
#define KEY_SF4		256+87
#define KEY_SF5		256+88
#define KEY_SF6		256+89
#define KEY_SF7		256+90
#define KEY_SF8		256+91
#define KEY_SF9		256+92
#define KEY_SF10	256+93
#define KEY_SF11	256+135
#define KEY_SF12	256+136
#define KEY_CF1		256+94
#define KEY_CF2		256+95
#define KEY_CF3		256+96
#define KEY_CF4		256+97
#define KEY_CF5		256+98
#define KEY_CF6		256+99
#define KEY_CF7		256+100
#define KEY_CF8		256+101
#define KEY_CF9		256+102
#define KEY_CF10	256+103
#define KEY_CF11	256+137
#define KEY_CF12	256+138

#define KEY_HOME	256+71
#define KEY_AHOME	256+151
#define KEY_CHOME	256+119

#define KEY_UP		256+72
#define KEY_CUP		256+141
#define KEY_AUP		256+152

#define KEY_PGUP	256+73
#define KEY_CPGUP	256+132
#define KEY_APGUP	256+153

#define KEY_LEFT	256+75
#define KEY_CLEFT	256+115
#define KEY_ALEFT	256+155

#define KEY_RIGHT	256+77
#define KEY_CRIGHT	256+116
#define KEY_ARIGHT	256+157

#define KEY_END		256+79
#define KEY_CEND	256+117
#define KEY_AEND	256+159

#define KEY_DOWN	256+80
#define KEY_CDOWN	256+145
#define KEY_ADOWN	256+160

#define KEY_PGDN	256+81
#define KEY_CPGDN	256+118
#define KEY_APGDN	256+161

#define KEY_INS		256+82
#define KEY_CINS	256+146
#define KEY_AINS	256+162

#define KEY_DEL		256+83
#define KEY_CDEL	256+147
#define KEY_ADEL	256+163

#define KEY_CNUM5	256+143
#define KEY_SNUM5	256+76

#ifndef BYTE_TYPES
#define BYTE_TYPES
typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
#endif

typedef unsigned char bool;
#define false 0
#define true (!false)


#define MAX_STYLE 32

typedef struct trk_info_tag {
		char *fname;		// Path to file which can be passed to fopen
		char *title;
		char *artist;
		unsigned int len;	// length in seconds
		unsigned long last_played; // last time it was played
		int  styles[MAX_STYLE];
	} trk_info_t;

typedef struct trk_list_tag {
		WORD trk;
	} trk_list_t;

typedef struct play_list_tag {
		WORD trk_index;		// Look this up in the
		unsigned long time;	// time since 00:00 Jan 1st 1970
		unsigned int sequence;	// Used for playlist entries
		long score;		// Used for sorting
	} play_list_t;

/* In findfile.c */
int  track_init(void);
void track_exit(void);
int  track_add_at(unsigned trk,unsigned pos);
int  track_add(unsigned trk);
void track_delete_from(unsigned pos);
void track_delete_last(void);
int  track_find(int list);
unsigned track_valid(unsigned trk);
void track_sort(void);

/* In juke????.c */
void status_init(void);
void status_exit(void);
char *track_label(unsigned trk);
void status_refresh(void);
void status_update(void);
void status_update_playlist(void);

/* In lcd.c (used by jukereal.c) */
void lcd_init(void);
void lcd_exit(void);
void lcd_update(void);
void lcd_update_time(unsigned int seconds);

/* In hwint.c (used by jukereal.c) */
void  switch_init(void);
DWORD switch_read(void);
void  solenoid(int on);
void  lamp1(int on);
void  lamp2(int on);

/* In serial.c (used by disp_com.c) */
int  serial_init(void);
void serial_exit(void);
void serial_tx_clear(void);
bool serial_tx_empty(void);
void serial_tx(BYTE ch);
bool serial_rx_ready(void);
BYTE serial_rx(void);

/* In disp_???.c */
int  display_init(void);
void display_exit(void);
void display_str(char *str);
void display_clear(void);
void display_xy(int row,int col);
void display_row(int row);
void display_clr_eol(void);
int  display_getkey(void);
void display_playing_status(int secs,int busy);
void display_flush(void);
bool display_cancel(void);


/* In dosaudio.c */
int audioPlay(char *buffer,int nBytes);
int audioInit(int initRate,int initBits,int initChans);
void audioComplete(void);
void audioExit(void);
bool volume_write(int port,int val);

/* In plaympeg.c */
int player_init(void);
void player_abort_playing(void);
int player_play_file(char *fname);
void player_exit(void);
void write_to_log(char *str);


/* Defined in findfile.c */
extern play_list_t toplay[];
extern int ntoplay,max_toplay,total_played;
extern bool really_exit,background_mode;

extern trk_info_t *trks;
extern int ntrks,nsorted;
extern unsigned tracks_bya[], tracks_byt[];
extern trk_list_t trk_lists[];

#endif
