/* 
 * The physical jukebox control software
 * -------------------------------------
 * copyright (c) 1997-1999 by Roger Walker, All rights reserved.
 * See the file RNW_COPY
 *
 * Uses routines in hwint.c to work the physical hardware
 *
 *  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
 *  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
 *  PURPOSE.
 */
#include <conio.h>
#include <dos.h>
#include <string.h>
#include <memory.h>
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include "jukebox.h"

extern trk_info_t *trks;
extern int ntrks;

extern play_list_t toplay[];
extern int ntoplay,max_toplay;
extern bool really_exit,background_mode;
WORD secs_remaining;

/* Letters buttons are bits 0-19, not that the letters I and O are missing */
char letters[20]="ABCDEFGHJKLMNPQRSTUV";
#define NUM_LETTERS ((int)sizeof(letters))


// EXPORTED functions
#define PLAY_LIST_1 3
#define PLAY_LIST_2 6

// Not used
void run_idle_task(void)
{ static int initial=0;
  if (initial==0)
  { initial++;
    // player_play_file("end.mp3");    
  }
  status_update();
}

// When something has changed, update the LCD display
void display_update_status(void)
{ lcd_update();
}

// After a change in the number of lined up track update status
// Work out which lights to leave on
// -ve is a special case which always gives blank (we are about to exit)
void status_update_tracklist(void)
{ if (ntoplay<0)
  { lamp1(0); lamp2(0);
    lcd_exit();
    return;
  } else
  if ((max_toplay-ntoplay)>6)
  { lamp1(1); lamp2(1);
  } else
  if ((max_toplay-ntoplay)>3)
  { lamp1(1); lamp2(0);
  } else
  if ((max_toplay-ntoplay)>1)
  { lamp1(0); lamp2(1);
  } else
  { lamp1(0); lamp2(0);
  }
  lcd_update();
}

// Just for debugging !
void display_playing_status(int nSecs,int nBusy)
{ display_xy(70,0);
  printf("%3d:%02d %3d",nSecs/60,nSecs%60,nBusy);
  lcd_update_time(nSecs);
}

// Return a meaningfull track ident, used for log file entries
char *track_label(unsigned trk)
{ static char retbuf[3];
  if (trk>=NUM_LETTERS*NUM_LETTERS)
  { // Above 400 gets AA, AB etc.
    trk-=NUM_LETTERS*NUM_LETTERS;
    retbuf[0]=letters[trk/NUM_LETTERS];
    retbuf[1]=letters[trk%NUM_LETTERS];
  } else
  if ((trk%2)==0)
   { trk/=2;
     retbuf[0]=letters[trk%NUM_LETTERS];
     retbuf[1]='0'+(trk/NUM_LETTERS);
   } else
   { trk/=2;
     retbuf[0]='0'+(trk/NUM_LETTERS);
     retbuf[1]=letters[trk%NUM_LETTERS];
   }
  retbuf[2]='\0';
  return retbuf;
}

// Given a letter and number reference, calculate an index into the laded tracksd
static int track_num(int let_key,int num_key)
{ return (num_key-NUM_LETTERS)*NUM_LETTERS+let_key;
}

#define TRK_NOBUT -1
#define TRK_LBUT -2
#define TRK_NBUT -3
#define TRK_3BUT -4

// Given the button down DWORD, work out what track to add (-ve =err)
static int track_number(DWORD buttons)
{ int but,but1,but2; DWORD tst;

  /* No buttons found yet */
  but1=but2=0;

  /* Test for buttons (numbered 1-32 for bits 0-31) */
  for (but=1,tst=1 ; but<=32 ; but++,tst<<=1)
   if (buttons&tst)
    { /* Do we already have two buttons ? */
      if (but2) return TRK_3BUT;
      /* got the first ? */
      if (but1) but2=but;
      else but1=but;
    }

  /* Did we find no buttons ? */
  if (but1==0) return TRK_NOBUT;

  /* Did we only find 1 button ? */
  if (but2==0)
   { return (but1<=NUM_LETTERS) ? TRK_LBUT : TRK_NBUT;
   }

  /* Convert both buttons to 0 based */
  but1--; but2--;

  /* Do we have one button in each section */
  if (but2<NUM_LETTERS || but1>=NUM_LETTERS)
   return TRK_3BUT;

  /* Return the track as a value between 0-219 */
  return track_num(but1,but2);
}


/* Maximum of 60 second, polling at least every .75 */
#define TIME_MAX  6000
#define TIME_POLL 50
#define TIME_FLASH 100

/* Get time in 1/100th second units. Range 0-5999 (1 minute) */
static WORD current_time(void)
{ struct dostime_t time;
  _dos_gettime(&time);
  return (time.second*100L + time.hsecond)%TIME_MAX;
}

/* Find time remaining */
static WORD remaining_time(WORD target)
{ WORD current;
  current=current_time();
  return (target>current) ? (target-current) : (TIME_MAX+target-current);
}

/* Calculate the time at which to time out */
static WORD target_time(WORD timeout)
{ return (current_time()+timeout)%TIME_MAX;
}

WORD end_time;
typedef enum { b_none,b_one,b_two,b_wait } jmode_t;
jmode_t bmode; 
int to_add,first_down;

// Current status, used for debug strings
static char *s_stat(jmode_t m)
{ return (m==b_none) ?  "None" :
         (m==b_one) ?   "One" :
         (m==b_two) ?   "Two" :
         (m==b_wait) ?  "Wait" : "Unknown";
}

// State machine for change of mode
static void status_change(jmode_t new_mode)
{ int next_change;
  status_update_tracklist();

  next_change=0;
  switch( new_mode )
   { case b_none:
       /* Waiting for something to happen, so don't time out */
       solenoid(0);
       break;

     case b_one:
       // If already too many tracks then ignore
       if (ntoplay>=max_toplay) 
       { new_mode=b_none;
         solenoid(0);
         break;
       }
       /* One button is down, time out after 30 seconds */
       next_change=3000;
       /* Hold the button down when they remove their finger */
       solenoid(1);
       break;

     case b_two:
       // If already too many tracks then ignore
       if (ntoplay>=max_toplay) 
       { new_mode=b_none;
         solenoid(0);
         break;
       }
       /* Both buttons are down, wait for user to remove fingers */
       next_change=100;
       /* Should have come after b_one, but maybe not if they were quick */
       solenoid(1);
       break;

     case b_wait:
       /* To get here someone is holding the keys down ! */
       /* This is used for volume control etc. */
       /* Release the solenoid and wait for the keys to release */
       solenoid(0);
       break;
   }

  bmode=new_mode;
  end_time=target_time(next_change);

  display_xy(40,0);
  printf("Mode %-8s",s_stat(bmode));
}

// The master volume level
static int volume;

// Reset the internal states
void status_reset(void)
{ 
  display_clear();
  switch_init();
  status_update_tracklist();
  lcd_init();
//  init_output();

  /* Set status illegal buttons pressed */
  /* That way if a button is already down it will be released */
  status_change(b_wait);

  /* Set volumes to 90% */
  volume=VOL_MAXVAL*9/10;
  volume_write(VOL_WAVE,   volume*VOL_SCALE);
  volume_write(VOL_MASTER, volume*VOL_SCALE);

  /* Mute (unused) line and CD in */
  volume_write(VOL_LINE,0);
  volume_write(VOL_CD,0);

  /* Set tones to midway */
  volume_write(TONE_BASE,(TONE_MAXVAL*TONE_SCALE/2));
  volume_write(TONE_TREBLE,(TONE_MAXVAL*TONE_SCALE/2));
  max_toplay=12;
}

static WORD last_special=0;	// Debounces the special key sequences

// (Referred to by the LCD routines)
bool locked=false;	// Locks the special key sequences

// Check for a hidden key sequence (two or more non-adjacent numbers held down)
static void check_special_keys(DWORD buttons)
{ WORD numbers; int l,r,n,i,j;

  // Mask off the number buttons
  numbers=(buttons>>NUM_LETTERS)&((1<<10)-1);

  // Find the left and right number buttons which are held down
  for (i=0,l=r=0 ; i<9 ; i++)
  { for (j=i+1 ; j<10 ; j++)
     if (numbers==((1<<i)|(1<<j))) 
     { l=i; r=j; break;
     }
    if (r!=0) break;
  }
//display_xy(0,17); printf("%03X %2d%-2d %d",numbers,l,r,locked);

  // Check for lock / unlock keys
  if (l==0 && r==1)
  { // LOCK sequence
    locked=true;
    last_special=numbers;    
    lcd_update();
    return;
  } else
  if (l==2 && r==9)
  { // UNLOCK sequence
    locked=false;
    last_special=numbers;    
    lcd_update();
    status_update_tracklist();
    return;
  }

  // If locked and two number keys are held down, flash the lights
  if (locked)
  { BYTE lamps;
    lamps=(current_time()%100)/50;
    lamp1(lamps); lamp2(lamps);
    return;
  }

  // If not two numbers then ignore (either 1 or 3+)
  if (r==0) return;

  // Same as last time so ignore
  if (last_special==numbers) return;

  // Delete something
  if (l==0)
  { if (r==8)
    { track_delete_last();
    } else
    if (r==7)
    { // Limit number of tracks in play list
      if (ntoplay>PLAY_LIST_2) ntoplay=PLAY_LIST_2;
      lcd_update();
    } else
    if (r==6)
    { // Limit number of tracks in play list
      if (ntoplay>PLAY_LIST_1) ntoplay=PLAY_LIST_1;
      lcd_update();
    } else
    if (r==5)
    { // Limit number of tracks in play list
      if (ntoplay>1) ntoplay=1;
      lcd_update();
    } else
    if (r==3)
    { // Delete current track
      abort_playing();
    }
  } else

  // 3 & 9 will put it in play background music mode
  if (l==3 && r==9)
  { background_mode=true;
    status_update_tracklist();
    lcd_update();
  } else

  // 3 & 6 will start a sequence (and take it out of background mode)
  if (l==3 && r==6)
  { int l,t;
    static unsigned int sequence=0;

    background_mode=false;
    status_update_tracklist();
    lcd_update();

    // Mask off the letter key information
    buttons&=(1UL<<NUM_LETTERS)-1;

    for (n=0 ; n<NUM_LETTERS ; n++)
    { if (buttons==(1UL<<n)) break;
    }
    // Not a single button down (none or two+)
    if (n>=NUM_LETTERS) return;

    // Add track group here
    // Start with the first in the list
    l=1;
    // Search all tracks for it
    do for (t=0 ; t<ntrks ; t++)
        if (trks[t].styles[n]==l)
         { // When we find it add to list
           track_add(t);
           if (toplay[ntoplay-1].trk_index==t)
            { // Force place in playlist
              toplay[ntoplay-1].sequence= ++sequence;
            }
           // Search for the next one
           l++;
           break;
         }
    while (t<ntrks);  // If not found bail out
    // Done
  } else

  if (l==5 || r==5)
  { if (l==1)
    { /* Mute volume */
      volume_write(VOL_MASTER,0);
    } else
    if (l==3)
    { /* Un-Mute volume */
      volume_write(VOL_MASTER,volume*VOL_SCALE);
    } else
    if (r==7)
    { /* Decrease volume */
      if (volume>0) volume--;
      volume_write(VOL_MASTER,volume*VOL_SCALE);
    } else
    if (r==9)
    { /* Increase volume */
      if (volume<VOL_MAXVAL) volume++;
      volume_write(VOL_MASTER,volume*VOL_SCALE);
    }
  }
  if (l==1 || r==3)
  { lcd_init();
    lcd_update();
  }

  // If we didn't bail out then remember the current keys
  last_special=numbers;    
  return;
}


static int status_toplay=0;	// Used to track changes in number of tracks outstanding

// Force a full refresh
void status_refresh(void)
{ status_toplay= -1;
  status_update();
}

// Update what's changed
void status_update(void)
{ static int hit= -1;		// Use by the real keyboard emulation
  int trk; DWORD buttons;

  if (kbhit())
   { // DEBUG mode, has a real keyboard, so allow track entry from it
     int key; 
     key=getch();
     if (key==0) key=256+getch();
     else key=toupper(key);

     if (key>='A' && key<='V')
      { if (key>='O') key-=2; else
        if (key>='I') key--;
        key-='A';
        if (hit<NUM_LETTERS) /* Which includes -1 */
         { /* Use this keyhit as the letter */
           hit=key;
         } else
         { /* hit holds a number, so select track */
           track_add(track_valid(2*track_num(key,hit)+1));
           hit= -1;
         }
      } else
     if (key>='0' && key<='9')
      { key-='0';
        key+=NUM_LETTERS;
        if (hit== -1 || hit>=NUM_LETTERS) /* No hit or hit was number */
         { /* Use this keyhit as the number */
           hit=key;
         } else
         { /* hit holds a letter, so select track */
           track_add(track_valid(2*track_num(hit,key)));
           hit= -1;
         }
      } else
     switch (key)
      { case KEY_ESC:
        case KEY_CF12:
        case KEY_SF12:
         abort_playing();
         really_exit=true;
         break;
        case KEY_F10:
         background_mode^=true;
         break;
        case KEY_F11:
         track_delete_last();
         break;
        case KEY_F12:
         abort_playing();
         break;
      }
   }

  buttons=switch_read();
  trk=track_number(buttons);

  if (trk==TRK_3BUT && bmode!=b_wait)
   { /* For any mode which isn't b_wait
      * If illegal buttons are down, change to b_wait
      */
     status_change(b_wait);
     // Clear timer display
     display_xy(60,0); printf("    ");
   } else
  switch (bmode)
   { case b_none:
      if (trk==TRK_LBUT || trk==TRK_NBUT)
       { status_change(b_one);
         first_down=trk;
       }
      /* Drop into b_one tests */
     case b_one:
      if (trk>=0)
       { status_change(b_two);
       }
      break;

     case b_wait:
      if (trk==TRK_NOBUT)    
       { /* No buttons are down so reset mode */
         status_change(b_none);
         last_special= 0;
       } else
      if (trk==TRK_LBUT || trk==TRK_NBUT)
       { /* Only one button down now so key debounce */
         last_special= 0;
         // Call routine so that if locked we can flash the lights
         check_special_keys(buttons);
       } else
      if (trk>=0)
       { /* We now have a valid track, 
          * we might be waiting for the user to let go, or the buttons to pop out
          * or someone might be letting go of a three button code
          */
       } else
       { check_special_keys(buttons);
       }
      break;
   }

  /* If this mode has a time out, is it up ? */
  if (bmode==b_one || bmode==b_two)
   { int timrem;
     /* Set cursor up for time display or blanking */
     display_xy(60,0);
     
     /* Find time since it expired */
     timrem=remaining_time(end_time);

     /* If less than polling time then has expired */
     if (timrem<TIME_POLL)
       { if (bmode==b_two)
          { /* We have had two buttons for a valid amount of time */
            /* If track is valid add it */
            if (trk>=0)
             { trk*=2;
               if (first_down==TRK_NBUT) trk++;
               track_add(track_valid(trk));
             }
          }
         printf("    ");
         /* Turn the solenoid off */
         /* The buttons will return to none eventually */
         status_change(b_wait);
       } else
       { printf("%-4u",timrem);
       }
   }

  /* If number of tracks lined up has changed then update display */
  if (status_toplay!=ntoplay)
   { int i,j;
     status_update_tracklist();
     lcd_update();
     display_xy(0,1);
     printf("%d tracks lined up  \n",ntoplay);
     for (i=0 ; i<10 ; i++)
     { display_xy(0,2+i);
       if (i>=ntoplay)
       { printf("%80s","");
       } else
       { trk_info_t *tp;
         tp=trks+toplay[i].trk_index;
         printf("%5d %s %-35s %-45s",
                toplay[i].score,track_label(toplay[i].trk_index),tp->artist,tp->title);
       }
     }
     status_toplay=ntoplay;
   }

  /* Leave the cursor somewhere consistant */
  display_xy(0,18);
}
