/* 
 * Machine dependent library
 * --------------------------
 * copyright (c) 1987-1999 by Roger Walker, All rights reserved.
 * See the file RNW_COPY
 *
 * These are routines I've built up over the years
 * Recently ported to DJGPP, and removed unused functions
 *
 * Portions Based on DMAW.C  ver 1.01  (Aug 15, 94)
 * Copyright (C) 1994-96 Creative Technology.
 *
 *  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
 *  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
 *  PURPOSE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <dos.h>
#include "mlib.h"


#define DMA_MODE_MEM_TO_PORT    0x08
#define DMA_MODE_AUTOINIT       0x10
#define DMA_MODE_DECREMENT      0x20
#define DMA_MODE_SINGLE         0x40

#define DMA_MODE_VAL    (DMA_MODE_MEM_TO_PORT|DMA_MODE_AUTOINIT|DMA_MODE_SINGLE)

#define DMA8_FF_REG      0xC
#define DMA8_MASK_REG    0xA
#define DMA8_MODE_REG    0xB
#define DMA16_FF_REG    0xD8
#define DMA16_MASK_REG  0xD4
#define DMA16_MODE_REG  0xD6

#define DMA0_ADDR        0
#define DMA0_COUNT       1
#define DMA0_PAGE     0x87
#define DMA1_ADDR        2
#define DMA1_COUNT       3
#define DMA1_PAGE     0x83
#define DMA3_ADDR        6
#define DMA3_COUNT       7
#define DMA3_PAGE     0x82
#define DMA5_ADDR     0xC4
#define DMA5_COUNT    0xC6
#define DMA5_PAGE     0x8B
#define DMA6_ADDR     0xC8
#define DMA6_COUNT    0xCA
#define DMA6_PAGE     0x89
#define DMA7_ADDR     0xCC
#define DMA7_COUNT    0xCE
#define DMA7_PAGE     0x8A


/* Clear an area of the DMA output buffer (use 0x80, as in 8-bit = 0x80, 16-bit = 0x8080) */
void DmaClear(dma_memory_block_t *blk,WORD index,WORD count)
{
#ifdef __DJGPP__
   DWORD dest = blk->off32+index;
   _farsetsel(blk->sel32);
   while(count--) {
	  _farnspokeb(dest++, 0x80);
   }
#else
   memset(blk->pBase+index,0x80,count);
#endif
}


/* Copy data to an area of the DMA output buffer */
void DmaCopy(dma_memory_block_t *blk,DWORD index,DWORD count,char *src)
{
#ifdef __DJGPP__
   DWORD dest = blk->off32+index;
   _farsetsel(blk->sel32);
   while(count--) {
	  _farnspokeb(dest++, *(src++));
   }
#else
   memcpy(blk->pBase+index,src,count);
#endif
}

/* Find the position of the actual DMA hardware within the DMA buffer */
WORD DmaPos(dma_memory_block_t *blk)
{ long read1,read2;

  IRQ_DISABLE
  /* Set the flip-flop */
  outp((blk->is16bit) ? DMA16_FF_REG : DMA8_FF_REG, 0xFF);

  do 
  { /* Read twice */
	read1=inp(blk->adrCount); read1|=inp(blk->adrCount)<<8;
	read2=inp(blk->adrCount); read2|=inp(blk->adrCount)<<8;
	/* If too much of a gap, try again, because we rolled over */
	read1-=read2;
  } while (abs(read1)>64);

  IRQ_ENABLE

  /* If 16 bit channel convert words to bytes */
  if (blk->is16bit) read2*=2;

  /* Convert remaining bytes to position */
  return (blk->bufSize-read2)&0xfffc;
}


/* Allocate a DMA aligned block of DOS memory (ie. below 1M) */
BOOL DmaAlloc(dma_memory_block_t *blk,WORD size)
{	DWORD base;
    WORD toAlloc;

	/* Round up size and convert to number of paragraphs */
	toAlloc = (size + 15) >> 4;

	/* Save the paragraph rounded size */
	blk->bufSize=toAlloc<<4;

	/* Allocate twice as much as we actually need, so that we can align it on a DMA boundary */
	toAlloc = (toAlloc*2)-1;

/*	printf("%d paragraphs of memory = %dK\n",toAlloc,toAlloc*16/1024); */
#ifdef __DJGPP__
/* This function allocates DOS memory.  You pass it the number of
 * paragraphs ((bytes+15)>>4) to allocate.  If it succeeds, it returns a
 * segment (dos-style) and fills in _RET_SELECTOR_OR_MAX with a selector
 * (protected-mode) that you can use to reference the same memory.  Note
 * that it's the selector you use to free the block, not the segment.
 */
	base=__dpmi_allocate_dos_memory(toAlloc,&blk->sel32);
	/* Returned an illegal DOS style segment descriptor */
	if (base== 0xFFFFFFFFUL) return FALSE;
#else
	/* Check for block too big in DOS mode */
	if (size>=0x8000) return FALSE;
    blk->memBlock = (BYTE	*)malloc(toAlloc);
	/* Malloc failed */
    if (blk->memBlock==NULL) return FALSE;
#endif

	/* Convert DOS mode segment descriptor to physical address */
	base=(base&0xFFFF)<<4;

	/* Find a start which doesn't cross a DMA boundary */
	blk->base=base;
	while ( (blk->base>>16) != ((blk->base+size-1)>>16) ) 
           blk->base+=16;

#ifdef __DJGPP__
	/* Calculate the offset from the segment selector */
	blk->off32=blk->base-base;
#else
	/* Point the start start of the DMA buffer within the buffer */
	blk->pBase=blk->memBlock+(blk->base-base);
#endif

/* printf("Allocated DMA buffer at %X (32 bit %X:%X) length %X\n",retBlk.base,retBlk.sel32,retBlk.off32,size); */

	/* Initially clear the whole buffer */
	DmaClear(blk,0,blk->bufSize);

	return TRUE;
}

/* Free the memory allocated for the DMA buffer */
void DmaFree(dma_memory_block_t *blk)
{
#ifdef __DJGPP__
	__dpmi_free_dos_memory(blk->sel32);
#else
	free(blk->memBlock);
#endif
}

/*************************************************************************
*
* FUNCTION: InitDMADSP()
*
* DESCRIPTION: This function reads the first data block of the file and
*              from it obtains information that is needed to program the
*              DMA and DSP chips.  After reading the data block, the file
*              pointer points to the first byte of the voice data.
*
*              NOTE: The DMA chip is ALWAYS programmed for auto-init mode
*                    (command 0x58)!  The DSP chip will be programmed for
*                    auto-init or single-cycle mode depending upon
*                    conditions--see start_playback() for details.
*
*************************************************************************/
BOOL DmaInit(dma_memory_block_t *blk,int DMAChan)
{ WORD DMAAddr, DMACount, DMAPage, Offset, Page, Size;
  BYTE ByteTimeConstant;

  switch(DMAChan)
   { case 0:
        DMAAddr=DMA0_ADDR; DMACount=DMA0_COUNT; DMAPage=DMA0_PAGE; break;
     case 1:
        DMAAddr=DMA1_ADDR; DMACount=DMA1_COUNT; DMAPage=DMA1_PAGE; break;
     case 3:
        DMAAddr=DMA3_ADDR; DMACount=DMA3_COUNT; DMAPage=DMA3_PAGE; break;
     case 5:
        DMAAddr=DMA5_ADDR; DMACount=DMA5_COUNT; DMAPage=DMA5_PAGE; break;
     case 6:
        DMAAddr=DMA6_ADDR; DMACount=DMA6_COUNT; DMAPage=DMA6_PAGE; break;
     case 7:
        DMAAddr=DMA7_ADDR; DMACount=DMA7_COUNT; DMAPage=DMA7_PAGE; break;
     default:
        fputs("Bad DMA channel",stderr);
        return FALSE;
   }
  
  blk->channel=DMAChan;
  blk->adrDma=DMAAddr;
  blk->adrCount=DMACount;
  blk->is16bit=(DMAChan>=4) ? TRUE : FALSE;
  DMAChan%=4;
  

  /*--- PROGRAM THE DMA CHIP ---------------------------------------------*/
  /*----------------------------------------------------------------------*/
  Page   = (WORD) (blk->base >> 16);
  Offset = (WORD) (blk->base & 0xFFFF);

  if (blk->is16bit)
  {
    /* Offset for 16-bit DMA channel must be calculated differently... */
    /* Shift Offset 1 bit right, then copy LSB of Page to MSB of Offset. */
	Offset >>= 1;                       /* Divide Offset by 2 */
	if (Page & 0x0001) Offset|=0x8000;  /* Put LSB of Page into MSB of Offset */

	outp(DMA16_MASK_REG, DMAChan|4);    /* Disable DMA while prog. */
	outp(DMA16_FF_REG,   0);            /* Clear the flip-flop */

	outp(DMA16_MODE_REG, DMAChan|DMA_MODE_VAL); /* 16-bit auto-init */
	Size   = (blk->bufSize/2)-1;
	outp(DMACount, Size & 0xFF);        /* LO byte of count */
	outp(DMACount, Size >> 8);          /* HI byte of count */
  } else
  {
	outp(DMA8_MASK_REG, DMAChan|4);     /* Disable DMA while prog. */
	outp(DMA8_MODE_REG, DMAChan|DMA_MODE_VAL);  /* 8-bit auto-init */
	outp(DMA8_FF_REG, 0);               /* Clear the flip-flop */
	Size   = blk->bufSize - 1;
	outp(DMACount, Size & 0xFF);        /* LO byte of count */
	outp(DMACount, Size >> 8);          /* HI byte of count */
  }


  outp(DMAPage, Page);                  /* Physical page number */
  if (blk->is16bit==FALSE)   outp(DMA8_FF_REG, 0);               /* Clear the flip-flop */
  outp(DMAAddr, Offset & 0xFF);         /* LO byte address of buffer */
  outp(DMAAddr, Offset >> 8);           /* HI byte address of buffer */

  /* Done programming the DMA, enable it */
  outp((blk->is16bit==FALSE) ? DMA8_MASK_REG : DMA16_MASK_REG, DMAChan);
  return TRUE;
}


#define PIC_MASTER 0x20
#define INT_BASE_MASTER 0x08
#define PIC_SLAVE  0xA0
#define INT_BASE_SLAVE 0x70

#define PIC_END_OF_INT           0x20
#define PIC_MODE                    0
#define PIC_MASK                    1


/*************************************************************************
*
* FUNCTION: SetIrqHandler()
*
* DESCRIPTION: Points an interrupt vector to a new handler, returns the original
*
*************************************************************************/
static irq_save_t *IrqSetHandler(BYTE vector,irq_save_t *handler)
{	
	irq_save_t *oldvect;
#ifdef __DJGPP__
	_go32_dpmi_seginfo seginfo;

	/* Read original vector */
	_go32_dpmi_get_protected_mode_interrupt_vector(vector, &seginfo);
/* printf("Old handler at %04X:%08X\n",seginfo.pm_selector,seginfo.pm_offset); */
	oldvect = (irq_save_t *)seginfo.pm_offset;
	seginfo.pm_offset = (int)handler;
	seginfo.pm_selector = _go32_my_cs();
	_go32_dpmi_allocate_iret_wrapper(&seginfo);
	_go32_dpmi_set_protected_mode_interrupt_vector(vector, &seginfo);

#else
	/* REAL mode */
	oldvect = _dos_getvect(vector);
	_dos_setvect(vector,handler);
#endif

	return oldvect;
}


BOOL IrqInit(irq_info_t *irqInfo,int irq,irq_interrupt_t *intHandler)
{ BYTE IRQMask;
  irqInfo->PICbase=(irq<8)  ? PIC_MASTER : PIC_SLAVE;
  irqInfo->ISRvec =((irq<8) ? INT_BASE_MASTER : INT_BASE_SLAVE)+(irq%8);

  /*--- SAVE CURRENT ISR FOR IRQVector, THEN GIVE IRQNumber NEW ISR ---*/
  /*-------------------------------------------------------------------*/
  irqInfo->oldVector = IrqSetHandler(irqInfo->ISRvec,intHandler);

  /* Save old interrupt mask */
  irqInfo->oldMaskVal = inp(irqInfo->PICbase+PIC_MASK);

  /* Enable previous AND new interrupts */
  IRQMask = (BYTE)(1 << (irq%8)); /* Shift a 1 left irq of bits */
  outp(irqInfo->PICbase+PIC_MASK, (irqInfo->oldMaskVal & ~IRQMask));
}


void IrqAck(irq_info_t *irqInfo)
{ if (irqInfo->PICbase==PIC_SLAVE)
   { /* Clear master PIC first */
     outp(PIC_MASTER+PIC_MODE, PIC_END_OF_INT); /* End of interrupt */
   }
  outp(irqInfo->PICbase	+PIC_MODE, PIC_END_OF_INT); /* End of interrupt */
}


void IrqEnd(irq_info_t *irqInfo)
{ outp(irqInfo->PICbase+PIC_MASK, irqInfo->oldMaskVal);
  IrqSetHandler(irqInfo->ISRvec,irqInfo->oldVector);
}
