#ifndef __MLIB_H
#define __MLIB_H

/* DJGPP protected mode stuff */
#ifdef __DJGPP__
#include <dpmi.h>

/* DJGPP ignored prefixes */
#define _far
#define _interrupt
#define _cdecl

/* DJGPP redefine */
#define ltoa itoa
#endif /* __DJGPP__ */

/* MSC versions < 8.0 don't know about inline */
#if _MSC_VER<=800
#define inline static
#endif


/* Standard varible types */
#ifndef BYTE_TYPES
#define BYTE_TYPES
typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
#endif

typedef unsigned char BOOL;
#define FALSE 0
#ifdef TRUE
#undef TRUE
#endif
#define TRUE (!FALSE)

#ifdef __DJGPP__
/* DOS Protected mode DMA support */
typedef struct
{	DWORD	base;			/* The physical base address of the aligned block */
	BOOL	is16bit;		/* True if using a 16 bit channel */
	BYTE	channel;		/* The DMA channel in use */
	WORD	bufSize;		/* The size of the DMA buffer */
	WORD	adrDma;			/* The base address of the DMA registers */
	WORD	adrCount;		/* The address of the DMA count register */

	DWORD	off32;			/* The offset from the segment selector for the memory block */
	int		sel32;       	/* The segment selector which describes it */
} dma_memory_block_t;

#else
/* DOS, real mode DMA support */
typedef struct
{	DWORD	base;			/* The physical base address of the aligned block */
	BOOL	is16bit;		/* True if using a 16 bit channel */
	BYTE	channel;		/* The DMA channel in use */
	WORD	bufSize;		/* The size of the DMA buffer */
	WORD	adrDma;			/* The base address of the DMA registers */
	WORD	adrCount;		/* The address of the DMA count register */

	BYTE	*pBase;			/* A pointer to the start of the DMA buffer within the memory block */
	BYTE	*memBlock;		/* The block of memory we allocated to get it */
} dma_memory_block_t;

#endif

BOOL DmaAlloc(dma_memory_block_t *blk,WORD size);
void DmaFree(dma_memory_block_t *blk);
BOOL DmaInit(dma_memory_block_t *blk,int channel);
void DmaClear(dma_memory_block_t *blk,WORD index,WORD count);
void DmaCopy(dma_memory_block_t *blk,DWORD index,DWORD count,char *src);
WORD DmaPos(dma_memory_block_t *blk);


typedef void _interrupt _far irq_interrupt_t;
typedef irq_interrupt_t (irq_save_t)(void);

typedef struct
{	WORD	PICbase;
	WORD	ISRvec;
	irq_save_t *oldVector;
	BYTE	oldMaskVal;
} irq_info_t;

BOOL IrqInit(irq_info_t *irqInfo,int irq,irq_interrupt_t *intHandler);
void IrqAck(irq_info_t *irqInfo);
void IrqEnd(irq_info_t *irqInfo);

int  audioInit(int initRate,int initBits,int initChans);
int  audioPlay(char *buffer,int nBytes);
void audioComplete(void);
void audioExit(void);

#ifdef __DJGPP__
#define IRQ_DISABLE	__asm__(" pushf\n sti");
#define IRQ_ENABLE	__asm__(" popf");
#else
#define IRQ_DISABLE _asm pushf; _asm sti;
#define IRQ_ENABLE	_asm popf;
#endif


#endif /* __MLIB_H */
